# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implements a single API for instrumenting all dbapi2-compliant modules
"""
import contrast
from contrast.api.architecture_component import ArchitectureComponent
from contrast.applies.sqli import apply_rule
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    pack_self,
    wrap_and_watermark,
)


class Dbapi2Patcher(object):
    """
    Do not instantiate this class; call `instrument_adapter` or subclass it.

    This class provides machinery to instrument a totally generic, PEP-249 compliant
    adapter. We only have a reference to the adapter module, and we can't make any
    assumptions about the existence of `adapter.Cursor`, since this is not guaranteed
    by the spec.

    We are only guaranteed the following:
    - the adapter has a `connection()` method, which returns an instance of Connection
    - the Connection object has a `cursor()` method, which returns an instance of Cursor
    - the Cursor has `execute()` and `executemany()` methods

    This requires a somewhat roundabout instrumentation strategy:
    - on the first call to adapter.connect(), we can instrument the Connection class
    - on the first call to Connection.cursor(), we can instrument the Cursor class
    - this lets us instrument Cursor.execute() and Cursor.executemany()
    """

    def __init__(self, adapter, vendor):
        self.dbinfo = {"vendor": vendor}
        self.vendor = vendor

        self._adapter = adapter
        # this module name must match a policy node; we need to fake sqlalchemy
        self._adapter_name = (
            "sqlalchemy" if self.vendor == "sqlalchemy" else self._adapter.__name__
        )
        self._connect_called = False
        self._cursor_called = False

    @fail_quietly("failed to instrument database adapter")
    def instrument(self, extra_cursors):
        build_and_apply_patch(self._adapter, "connect", self._build_connect_patch)
        for cursor in extra_cursors:
            self._safe_instrument_cursor(cursor)

    @fail_quietly("failed to get database inventory information")
    def init_dbinfo(self, connection, connect_args, connect_kwargs):
        """
        This method is intended to be overridden by subclasses if necessary. It must add
        key-value pairs to `dbinfo` - any/all of host", "port", and "database" (ie
        dbname). "vendor" should already be populated.

        Takes the arguments from a call to connect() and extracts information for
        database inventory reporting. By default only works with kwargs - we expect this
        to be by far the most common case.

        This follows the conventions laid out in the footnotes of dbapi2:
        https://www.python.org/dev/peps/pep-0249/#footnote-1

        In SQLAlchemy, see create_connect_args() for each dialect to get an idea of how
        different adapters are used. Most, if not all, use kwargs. Some adapters support
        a DSN connection string, but that should be handled by subclasses (if at all).
        """
        for key in ["host", "port"]:
            if key in connect_kwargs:
                self.dbinfo[key] = connect_kwargs[key]

        # the database name argument has several variations
        for key in ["database", "dbname", "db"]:
            if key in connect_kwargs:
                self.dbinfo["database"] = connect_kwargs[key]
                break

    @fail_quietly("failed to instrument database cursor class")
    def _safe_instrument_cursor(self, cursor_class):
        """
        Instruments a dbapi2-compliant database cursor class

        @param cursor_class: Reference to cursor class to be instrumented
        """
        self._instrument_cursor_method(cursor_class, "execute")
        self._instrument_cursor_method(cursor_class, "executemany")
        if hasattr(cursor_class, "executescript"):
            # non-standard, but provided by some adapters such as sqlite3
            self._instrument_cursor_method(cursor_class, "executescript")

    def _instrument_cursor_method(self, cursor, method_name):
        build_and_apply_patch(
            cursor,
            method_name,
            self._build_execute_patch,
        )

    @fail_quietly("failed to instrument database connection object")
    def _safe_instrument_connection(self, connection_instance):
        """
        Instruments a dbapi2-compliant database connection class, given an instance

        @param connection_instance: dbapi2 Connection instance
        """
        connection_class = type(connection_instance)
        build_and_apply_patch(connection_class, "cursor", self._build_cursor_patch)

    @property
    def _build_execute_patch(self):
        """
        See `_build_connect_patch`
        """

        def build_execute_patch(orig_func, _):
            def patched_method(wrapper, instance, args, kwargs):
                """
                Patch for dbapi_adapter.connection().cursor().execute*()
                """
                self._safe_create_arch_component_and_append_to_context()
                return apply_rule(
                    self._adapter_name,
                    wrapper,
                    pack_self(instance, args),
                    kwargs,
                )

            return wrap_and_watermark(orig_func, patched_method)

        return build_execute_patch

    @property
    def _build_cursor_patch(self):
        """
        See `_build_connect_patch`
        """

        def build_cursor_patch(orig_func, _):
            def cursor_patch(wrapped, instance, args, kwargs):
                """
                Patch for dbapi_adapter.connection().cursor()

                This patch will ensure that the returned Cursor object's class will have
                `execute` and `executemany` instrumented.
                """
                del instance

                cursor = wrapped(*args, **kwargs)
                if not self._cursor_called:
                    try:
                        cursor_class = type(cursor)
                        self._safe_instrument_cursor(cursor_class)
                        self._cursor_called = True
                    except Exception:
                        pass
                return cursor

            return wrap_and_watermark(orig_func, cursor_patch)

        return build_cursor_patch

    @property
    def _build_connect_patch(self):
        """
        Getter for `build_connect_patch`. We can't just make `build_connect_patch` an
        instance method because of the added `self` argument.
        """

        def build_connect_patch(orig_func, _):
            def connect_patch(wrapped, instance, args, kwargs):
                """
                Patch for dbapi_adapter.connection()

                This patch will ensure that the returned Connection object's class will
                have `cursor_patch` applied to its cursor() method.
                """
                del instance

                connection = wrapped(*args, **kwargs)
                if not self._connect_called:
                    self.init_dbinfo(connection, args, kwargs)
                    self._safe_instrument_connection(connection)
                    self._connect_called = True
                self._safe_create_arch_component_and_append_to_context()
                return connection

            return wrap_and_watermark(orig_func, connect_patch)

        return build_connect_patch

    @fail_quietly("failed to add database information to context")
    def _safe_create_arch_component_and_append_to_context(self):
        """
        Create ArchComponent with the given database information and append to context.database_info
        """
        context = contrast.CS__CONTEXT_TRACKER.current()
        if not context:
            return
        arch_comp = ArchitectureComponent(self.dbinfo).to_json()
        context.database_info.append(arch_comp)


@fail_quietly("failed to instrument database adapter")
def instrument_adapter(adapter, vendor, patcher_cls=Dbapi2Patcher, extra_cursors=None):
    """
    Instrument the provided dbapi2 adapter.

    `vendor` must be a string that exactly matches a value from Teamserver's
    flowmap/technologies.json > service > one of "name".

    References to cursors to instrument explicitly can be passed in via extra_cursors.
    If we could guarantee patches were applied before any calls to `adapter.connect()`,
    we wouldn't need to directly patch extra cursors - unfortunately, this probably
    means we'd need to be a runner.
    """
    patcher_cls(adapter, vendor).instrument(extra_cursors or [])



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re


HTTP_START = "http:"
HTTPS_START = "https:"
PERIOD = "."
SLASH = "/"


class EntityWrapper(object):
    SYSTEM_ID_REGEX = re.compile(r'<!ENTITY\s+([a-zA-Z0-f]+)\s+SYSTEM\s+"(.*?)">')
    PUBLIC_ID_REGEX = re.compile(
        r'<!ENTITY\s+([a-zA-Z0-f]+)\s+PUBLIC\s+".*?"\s+"(.*?)">'
    )
    DTD_MARKER = ".dtd"
    FILE_START = "file:"
    FTP_START = "ftp:"
    GOPHER_START = "gopher:"
    JAR_START = "jar:"
    UP_DIR_LINUX = "../"
    UP_DIR_WIN = "..\\"

    FILE_PATTERN_WINDOWS = re.compile(r"^[\\\\]*[a-z]{1,3}:.*", flags=re.IGNORECASE)

    def __init__(self, entity):
        self.system_id = self.parse_system_id(entity)
        self.public_id = self.parse_public_id(entity)

    def is_external_entity(self):
        if self.system_id:
            return self.external_id(self.system_id)
        if self.public_id:
            return self.external_id(self.public_id)

        return False

    def parse_system_id(self, entity):
        match = self.SYSTEM_ID_REGEX.search(entity)
        if not match:
            return ""
        return match.group(2)

    def parse_public_id(self, entity):
        match = self.PUBLIC_ID_REGEX.search(entity)
        if not match:
            return ""
        return match.group(2)

    def external_id(self, entity_id):
        tmp_id = entity_id.lower()

        http = self._starts_with_any(tmp_id, HTTP_START, HTTPS_START)

        if http and tmp_id.endswith(self.DTD_MARKER):
            return True

        if self._starts_with_any(
            tmp_id, self.FTP_START, self.FILE_START, self.JAR_START, self.GOPHER_START
        ):
            return True

        if self._starts_with_any(tmp_id, SLASH, PERIOD):
            return True

        if self._starts_with_any(tmp_id, self.UP_DIR_LINUX, self.UP_DIR_WIN):
            return True

        if self.FILE_PATTERN_WINDOWS.search(tmp_id):
            return True

        return False

    def _starts_with_any(self, string, *args):
        for arg in args:
            if string.startswith(arg):
                return True
        return False



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.response.base_body_only_rule import BaseBodyOnlyRule


class AutocompleteMissingRule(BaseBodyOnlyRule):
    @property
    def name(self):
        return "autocomplete-missing"

    def is_body_violated(self, body, form_tags):
        """
        Rule is violated if:
        1. none of the attrs of the list of form tags have an "autocomplete" attr
        2. at least one attr is "autocomplete" but it is is assigned to anything other than "off"
        :param body: response body
        :param form_tags: list of Tag namedtuple
        :return: bool, properties dict
        """
        if not form_tags:
            return False, {}

        for tag in form_tags:
            target_attr_found = False

            for attr_name, attr_value in tag.attrs:
                if attr_name.lower() == "autocomplete":
                    target_attr_found = True
                    if attr_value.lower() != "off":
                        properties = self.build_properties(tag.tag, body)
                        return True, properties

            # Report if no autocomplete attr found in the current form tag
            if not target_attr_found:
                properties = self.build_properties(tag.tag, body)
                return True, properties

        return False, {}



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsServerMessage
from contrast.agent.settings import Settings
from contrast.utils.decorators import fail_loudly
from contrast.utils.timer import now_ms


class ServerActivity(BaseTsServerMessage):
    def __init__(self):
        super().__init__()

        self.body = {"lastUpdate": self.since_last_update}

    @property
    def name(self):
        return "server-activity"

    @property
    def path(self):
        return "activity/server"

    @property
    def request_method(self):
        return requests.put

    @property
    def expected_response_codes(self):
        return [200, 304]

    @property
    def since_last_update(self):
        """
        Time in ms since server have been updated.
        If never updated, then it's been 0ms since then.
        """
        if self.settings.last_server_update_time_ms == 0:
            return 0
        return now_ms() - self.settings.last_server_update_time_ms

    @fail_loudly("Failed to process ServerActivity response")
    def process_response(self, response, reporting_client):
        settings = Settings()

        # TS will not send server settings unless lastUpdate is >= 5 mins (300_000 ms)
        # The response in those cases is 304.
        if (
            not self.process_response_code(response, reporting_client)
            or response.status_code == 304
        ):
            return

        body = response.json()

        settings.apply_ts_feature_settings(body)
        settings.process_ts_reactions(body)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import functools

import contrast
from contrast.agent import agent_state, scope as scope_
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.agent.asgi import (
    track_scope_sources,
    ASGIRequest,
    ASGIResponse,
)
from contrast.utils.decorators import cached_property
from contrast.utils.safe_import import safe_import_list
from contrast.utils.assess.duck_utils import safe_getattr_list

from contrast.utils import Profiler
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")

DEFAULT_ASGI_NAME = "asgi_app"


@functools.lru_cache(maxsize=1)
def _get_quart_types():
    return tuple(safe_import_list("quart.Quart"))


class ASGIMiddleware(BaseMiddleware):
    @scope_.with_contrast_scope
    def __init__(self, app, app_name=None, original_app=None):
        """
        Contrast middleware - ASGI 3.0

        The `app` argument here must be named "app" exactly to support FastAPI's
        middleware system, which uses the kwarg when calling this method.

        We need to keep the `original_app` kwarg for now to prevent a breaking API
        change.
        """
        del original_app

        # TODO: PYT-2852 Revisit application name detection
        self.app_name = (
            app_name
            if app_name is not None
            else safe_getattr_list(
                app,
                [
                    "__name__",
                    "name",
                    "title",
                ],
                DEFAULT_ASGI_NAME,
            )
        )

        super().__init__()

        if not agent_state.in_automatic_middleware() and isinstance(
            app, _get_quart_types()
        ):
            # we need this to prevent a breaking API change when wrapping Quart apps
            app = app.asgi_app

        self.asgi_app = app

    async def __call__(self, scope, receive, send) -> None:
        if contrast.CS__CONTEXT_TRACKER.current():
            # This can happen if a single app is wrapped by multiple instances of the
            # middleware (usually caused by automatic instrumentation)
            logger.debug("Detected preexisting request context")
            await self.asgi_app(scope, receive, send)
            return

        if scope.get("type") != "http":
            await self.call_without_agent_async(scope, receive, send)
        else:
            self.request_path = scope.get("path", "")
            with Profiler(self.request_path):
                request = ASGIRequest(scope, receive)
                environ = await request.to_wsgi_environ()

                context = self.should_analyze_request(environ)
                if context:
                    with contrast.CS__CONTEXT_TRACKER.lifespan(context):
                        await self.call_with_agent(context, request, send)
                        return

                await self.call_without_agent_async(scope, receive, send)

    async def call_with_agent(self, context, request, send) -> None:
        self.log_start_request_analysis()
        track_scope_sources(context, request.scope)

        try:
            self.prefilter()

            response = ASGIResponse(send)
            await self.asgi_app(
                request.scope, request.contrast__receive, response.contrast__send
            )

            context.extract_response_to_context(response)

            self.postfilter(context)
            self.check_for_blocked(context)

            await response.call_send()

        finally:
            self.handle_ensure(context, context.request)
            self.log_end_request_analysis()
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()

    async def call_without_agent_async(self, scope, receive, send) -> None:
        super().call_without_agent()
        with scope_.contrast_scope():
            await self.asgi_app(scope, receive, send)

    @cached_property
    def trigger_node(self):
        """
        trigger node used by reflected xss postfilter rule
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.asgi_app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    @cached_property
    def name(self):
        return "asgi"

    def get_route_coverage(self):
        """
        ASGI has no concept of "routes", so we can't perform route discovery without a
        framework
        """
        return {}



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor.wrapt import register_post_import_hook

from contrast.patches.databases import dbapi2

MYSQL_CONNECTOR = "mysql.connector"
VENDOR = "MySQL"


def instrument_mysql_connector(mysql_connector):
    extra_cursors = [
        mysql_connector.cursor.CursorBase,
        mysql_connector.cursor.MySQLCursor,
    ]

    try:
        # The C extension is technically optional
        extra_cursors.append(mysql_connector.cursor_cext.CMySQLCursor)
    except AttributeError:
        pass

    dbapi2.instrument_adapter(
        mysql_connector,
        VENDOR,
        dbapi2.Dbapi2Patcher,
        extra_cursors=extra_cursors,
    )


def register_patches():
    register_post_import_hook(instrument_mysql_connector, MYSQL_CONNECTOR)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast_vendor.wrapt import register_post_import_hook

from contrast.agent.policy.applicator import apply_module_patches
from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark

import urllib.parse as urllib

MODULE_NAME = "urllib.parse"


def build_urlencode_patch(orig_func, _):
    def urlencode_patch(wrapped, _, args, kwargs):
        """
        Patch for urllib.urlencode / urllib.parse.urlencode.

        This is needed because of the unfortunate way urllib calls on
        `quote_via` as a kwarg.
        See https://github.com/python/cpython/blob/master/Lib/urllib/parse.py#L909

        Because of this behavior, the `quote_via` func is not our patched
        `quote_via` defined in policy.
        By patching `urlencode`, we intercept its call and instead of allowing
        it to use the default kwarg for `quote_via`, we pass our own
        patched `quote_via` in order to prevent false positive XSS findings.
        """
        # quote_plus MUST already be patched by policy in order
        # to pass in the patched func to urlencode
        kwargs.setdefault("quote_via", urllib.quote_plus)
        return wrapped(*args, **kwargs)

    return wrap_and_watermark(orig_func, urlencode_patch)


def patch_urllib(urllib_module):
    # We ask policy to go ahead and do all urllib patches here (even though policy
    # patches will happen later on) because we MUST have some urllib policy patches
    # already applied for these non-policy patches to work.
    # This would not be necessary if in _enable_patches policy_patches were applied
    # first.
    apply_module_patches(urllib_module)

    build_and_apply_patch(urllib_module, "urlencode", build_urlencode_patch, force=True)


def register_patches():
    register_post_import_hook(patch_urllib, MODULE_NAME)


def reverse_patches():
    urllib_module = sys.modules.get(MODULE_NAME)
    if not urllib_module:  # pragma: no cover
        return

    patch_manager.reverse_patches_by_owner(urllib_module)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re
from contrast.agent.assess.rules.response.base_header_only_rule import (
    BaseHeaderOnlyRule,
)


class HstsHeaderRule(BaseHeaderOnlyRule):
    """
    Rule is violated if the header Strict-Transport-Security `max-age=` is present
    and assigned to 0.

    This is safe by default if the header isn't present.
    """

    @property
    def name(self):
        return "hsts-header-missing"

    @property
    def header_key(self):
        return "Strict-Transport-Security"

    @property
    def good_value(self):
        # unused
        pass

    @property
    def bad_value(self):
        return 0

    @property
    def good_values(self):
        # unused
        pass

    def is_header_violated(self, headers):
        value = headers.get(self.header_key)

        if value is None:
            # no finding if header is missing (safe by default)
            return False, None

        try:
            max_age_value = int(re.search(r"(?<=max-age=)\d+", value).group())
        except Exception:
            # If we know the header is there but for some reason can't process
            # the max age value, let's call it unsafe.
            properties = dict(
                type="Header",
                name=self.header_key,
                value=value,
            )
            return True, properties

        if max_age_value <= self.bad_value:
            properties = dict(
                type="Header",
                name=self.header_key,
                value=max_age_value,
            )
            return True, properties

        return False, None



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ArchitectureComponent(object):
    def __init__(self, db_info: dict):
        self.db_info = db_info
        self.type = "db"
        self.url = (
            db_info.get("database")
            if db_info.get("database")
            else (db_info.get("vendor") or "default")
        )

    def to_json(self):
        json = {
            "type": self.type,
            "url": self.db_info.get("database")
            if self.db_info.get("database")
            else (self.db_info.get("vendor") or "default"),
        }

        # vendor must be a string that exactly matches a value from
        # Teamserver's flowmap/technologies.json > service > one of "name"
        if self.db_info.get("vendor"):
            json["vendor"] = self.db_info.get("vendor")

        if self.db_info.get("host"):
            json["remoteHost"] = self.db_info.get("host")

        if self.db_info.get("port"):
            port = self.db_info.get("port")
            try:
                json["remotePort"] = str(int(port))
            except ValueError:
                # don't report port if it's not an int
                pass

        return json



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api.trace_event import TraceTaintRange
from contrast.utils.assess.tag_utils import merge_tags
from contrast.agent.assess.tag import Tag


ELLIPSIS = "..."
TRUNCATION_FRAGMENT_LENGTH = 50
TRUNCATION_LENGTH = TRUNCATION_FRAGMENT_LENGTH * 2 + len(ELLIPSIS)


def truncate(str_representation, length_factor=1):
    """
    Truncate + return the provided string
    """
    fragment_length = TRUNCATION_FRAGMENT_LENGTH * length_factor
    if len(str_representation) <= 2 * fragment_length + len(ELLIPSIS):
        return str_representation
    return "".join(
        [
            str_representation[:fragment_length],
            ELLIPSIS,
            str_representation[-fragment_length:],
        ]
    )


def truncate_tainted_string(str_representation, event):
    """
    Truncate + return the provided string and update the provided event's taint_ranges
    accordingly.

    This function assumes:
    - event.taint_ranges has already been calculated
    - each range in event.taint_ranges follows the format "{}:{}"
    - str_representation is the string representation of the tainted object associated
      with the given event, i.e. event.taint_ranges describes str_representation
    """
    if len(str_representation) <= TRUNCATION_LENGTH:
        return str_representation
    if len(event.taint_ranges) == 0:
        return truncate(str_representation)

    # In the eyes of TS, all tags are equally important for taint range calculation.
    # We take advantage of this here to maximally merge and simplify all tags
    # (regardless of tag key).
    merged_tags = _parse_and_merge_tags(event.taint_ranges)

    truncated, truncated_ranges = _truncate_tainted_string(
        str_representation, merged_tags
    )

    _update_event_taint_ranges(event, truncated_ranges)

    return truncated


def _parse_and_merge_tags(taint_range_objs):
    """
    Given a TraceEvent, parse out the existing tag ranges and merge them.

    merge_tags expects a dict, and merges based on tag key (dict key). We use a
    temporary key here for convenience; critically, all tags are given the same key so
    that they're merged together for TS taint calculations.
    """
    tags = {"temp": []}
    for taint_range_obj in taint_range_objs:
        start_idx, end_idx = (int(s) for s in taint_range_obj.range.split(":"))
        tags["temp"].append(Tag(end_idx - start_idx, start_idx))
    merge_tags(tags)
    return tags["temp"]


def _truncate_tainted_string(str_representation, merged_tags):
    """
    The core logic of tainted string truncation. Given a string and a list of simplified
    and merged Tags, truncate the string approximately according to this specification:

    https://bitbucket.org/contrastsecurity/assess-specifications/src/master
    filename: truncate-event-snapshots.md

    Note: this follows the specification very closely, but not exactly.
    """
    truncated = ""
    truncated_ranges = []
    curr_index = 0
    for tag in merged_tags:
        if curr_index < tag.start_index:
            truncated += truncate(str_representation[curr_index : tag.start_index])
        new_fragment = truncate(
            str_representation[tag.start_index : tag.end_index], length_factor=2
        )
        truncated += new_fragment
        truncated_ranges.append((len(truncated) - len(new_fragment), len(truncated)))
        curr_index = tag.end_index
    if curr_index < len(str_representation):
        truncated += truncate(str_representation[curr_index:])

    return truncated, truncated_ranges


def _update_event_taint_ranges(event, truncated_ranges):
    """
    Update a TraceEvent's taint_ranges according to the provided list of ranges.
    """
    event.taint_ranges.clear()

    for start_idx, end_idx in truncated_ranges:
        taint_range = TraceTaintRange()
        # this is not a "real" tag; TS only cares about ranges. The overall ranges
        # produced by this method are correct, but we drop individual tag ranges in
        # favor of a vastly simplified truncation algorithm.
        taint_range.tag = "UNTRUSTED"
        taint_range.range = f"{start_idx}:{end_idx}"
        event.taint_ranges.append(taint_range)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import copy
from contrast.api import Route
from contrast.utils.decorators import fail_quietly

from contrast.agent.middlewares.route_coverage.common import (
    DEFAULT_ROUTE_METHODS,
    build_args_from_function,
    build_key,
    get_normalized_uri,
)

DEFAULT_ROUTE_METHODS = copy.copy(DEFAULT_ROUTE_METHODS) + ("PUT", "PATCH", "DELETE")


def create_falcon_routes(app):
    """
    Given a Falcon app instance, use the private router
    to find all register routes. At this time, Falcon
    does not have a public API to get the app's routes.

    Borrowed from: https://stackoverflow.com/a/54510794

    :param app: class falcon.API or class falcon.APP instance
    :return: dict {route_id:  api.Route
    """
    routes = {}

    def get_children(node):
        if len(node.children):
            for child_node in node.children:
                get_children(child_node)
        else:
            create_routes(node.resource, node.uri_template, routes)

    for node in app._router._roots:
        get_children(node)

    return routes


def create_routes(endpoint_cls, path, routes):
    """
    Add to routes new items representing view functions for
    falcon class endpoint.

    :param endpoint_cls: Falcon class that defines views
    :param path: string of url path such as /home
    :param routes: dict of routes
    :return: None, side effect appends to routes
    """
    for method in DEFAULT_ROUTE_METHODS:
        view_func = get_view_method(endpoint_cls, method)

        if view_func:
            route = build_falcon_route(view_func, endpoint_cls)

            route_id = build_key(str(id(view_func)), method)
            routes[route_id] = Route(
                verb=method,
                url=get_normalized_uri(str(path)),
                route=route,
            )


def get_view_method(cls_instance, request_method):
    """
    Falcon defines views like this
    ```
    class Cmdi(object):
        def on_get(self, request, response):
            response.status = falcon.HTTP_200
            response.body = "Result from CMDI"
    ```
    Given this class definition and the request_method string,
    we will look for the correct view method.

    Note that we need to get the unbound method because later on we
    will get the id of this method.

    :param cls_instance: instance of Falcon class endpoint such as Cmdi in the above example
    :param request_method: string such as GET or POST
    :return: function: view method for the request_method, such as on_get for example above
    """
    view_name = f"on_{request_method.lower()}"
    if hasattr(cls_instance, view_name):
        # use .__class__ and/or __func__ to get the unbound method
        view_func = getattr(cls_instance.__class__, view_name)

        return view_func

    return None


def build_falcon_route(view_func, endpoint_cls):
    route_args = build_args_from_function(view_func)
    route = f"{endpoint_cls.__class__.__name__}.{view_func.__name__}{route_args}"
    return route


@fail_quietly("Unable to get route coverage", return_value={})
def get_route_coverage(falcon_app):
    return create_falcon_routes(falcon_app)


@fail_quietly("Unable to get Falcon view func")
def get_view_func(request_path, falcon_app, request_method):
    if not request_path:
        return None

    route_info = falcon_app._router.find(request_path)
    if not route_info:
        return None

    endpoint_cls, _, _, _ = route_info
    view_func = get_view_method(endpoint_cls, request_method)
    return view_func


@fail_quietly("Unable to build route", return_value="")
def build_route(view_func, url, falcon_app, request_path):
    route_info = falcon_app._router.find(request_path)
    if not route_info:
        return None

    endpoint_cls, _, _, _ = route_info
    return build_falcon_route(view_func, endpoint_cls)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsServerMessage
from contrast import get_canonical_version


class ServerStartup(BaseTsServerMessage):
    def __init__(self):
        super().__init__()

        self.body = {
            "environment": self.settings.config.get("server.environment"),
            "tags": self.settings.config.get("server.tags"),
            "version": get_canonical_version(),
        }

    @property
    def name(self):
        return "server-startup"

    @property
    def path(self):
        return "servers/"

    @property
    def expected_response_codes(self):
        return [200]

    @property
    def request_method(self):
        return requests.put

    @property
    def disable_agent_on_401_and_408(self):
        return True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import sys
import platform
import atexit
import threading
import warnings

from contextlib import contextmanager
from contextvars import ContextVar
from typing import Dict, Optional, Tuple

import contrast
from contrast import __version__, AGENT_CURR_WORKING_DIR
from contrast.utils.locale import DEFAULT_ENCODING
from contrast.agent import agent_lib, thread_watcher, patch_controller
from contrast.agent.assess.rules.providers.enable import enable_providers
from contrast.agent.settings import Settings
from contrast.api import Route
from contrast.reporting import ReportingClient, teamserver_messages
from contrast.reporting.reporting_client import MAX_ATTEMPTS
from contrast.assess_extensions import cs_str
from contrast.utils.exceptions.deprecation_warning import ContrastDeprecationWarning
from contrast.utils.loggers.logger import (
    setup_basic_agent_logger,
    setup_agent_logger,
    setup_security_logger,
)
from contrast.utils import timer
from contrast.utils.library_reader.library_reader import LibraryReader
from contrast.utils.namespace import Namespace
from contrast.utils.decorators import fail_loudly
from contrast_rewriter import (
    process_rewriter_logs,
    set_rewriter_logger,
)
from contrast.agent.runner import is_runner_in_use

# NOTE: policy is currently loaded/registered upon import
from contrast import policy  # pylint: disable=unused-import

logger = setup_basic_agent_logger()
LOGS_SEPARATOR = "-" * 120

SHARED_WARNING_MESSAGE = """
For specific circumstances where explicit middleware configuration is still
required, set `agent.python.enable_automatic_middleware` to `false` in the
configuration file or set `CONTRAST__AGENT__PYTHON__ENABLE_AUTOMATIC_MIDDLEWARE=false`
in your environment to suppress this warning.

Please see the documentation for additional details:
    https://docs.contrastsecurity.com/en/python-middleware.html
"""

MIDDLEWARE_WITHOUT_RUNNER_WARNING = f"""
Explicit middleware configuration is no longer required or recommended. Instead,
users should run their application with the `contrast-python-run` command, which
automatically detects and enables framework-specific middleware.
{SHARED_WARNING_MESSAGE}
"""

NO_EXPLICIT_MIDDLEWARE_WARNING = f"""
Explicit middleware configuration is generally no longer necessary when using
the `contrast-python-run` command.
{SHARED_WARNING_MESSAGE}
"""


NO_SPECIFIC_MIDDLEWARE_WARNING = """
Using framework-specific middleware is no longer recommended for {}.
Use contrast.{}.ContrastMiddleware instead.
"""


if not contrast.telemetry_disabled():
    from contrast.agent.telemetry import Telemetry
else:
    Telemetry = None


def stop_telemetry(telemetry):
    if telemetry.message_q:
        if not telemetry.message_q.empty():
            telemetry.send_messages()

        assert telemetry.message_q.empty()


class module(Namespace):
    init_lock: threading.Lock = threading.Lock()
    is_initialized = False
    id: Optional[int] = None
    settings: Optional[Settings] = None
    reporting_client: Optional[ReportingClient] = None
    library_reader: Optional[LibraryReader] = None
    routes: Dict[str, Route] = {}
    first_request: bool = True

    automatic_middleware: ContextVar = ContextVar("automatic_middleware", default=False)

    deprecated_middleware: Optional[Tuple[str, bool]] = None

    # NOTE: these field can be set prior to initialization
    framework: Optional[str] = None
    app_name: Optional[str] = None


def _log_environment(settings: Settings):
    """
    Log current working directory, python version and pip version
    """
    banner = f"{'-' * 50}ENVIRONMENT{'-' * 50}"
    logger.debug(banner)
    logger.debug("Current Working Dir: %s", AGENT_CURR_WORKING_DIR)
    logger.debug("Python Version: %s", sys.version)
    logger.debug("Detected Framework: %s", settings.framework)
    logger.debug("Detected application name: %s", module.app_name)
    logger.debug("Configured application name: %s", settings.app_name)
    logger.debug("Server Version: %s", settings.server)
    logger.debug("Contrast Python Agent Version: %s", __version__)
    logger.debug("Executable: %s", sys.executable)
    logger.debug("Default encoding %s", DEFAULT_ENCODING)
    logger.debug("Using runner: %s", is_runner_in_use())
    logger.debug("Has funchook: %s", cs_str.has_funchook())

    if sys.version_info[:2] == (3, 7):
        logger.warn(
            "Support for Python 3.7 is deprecated and will be removed in a future release"
        )

    try:
        platform_str = platform.platform()
    except Exception:
        try:
            platform_str = platform.platform(terse=True)
        except Exception:
            platform_str = "unknown"

    logger.debug("Platform %s", platform_str)

    try:
        import pip

        logger.debug("Pip Version: %s", pip.__version__)
    except Exception:
        logger.debug("Pip not found")

    logger.debug(banner)


def _warn_for_misleading_config(settings: Settings):
    protect_enabled = settings.is_protect_enabled()
    assess_enabled = settings.is_assess_enabled()

    logger.info("Protect: %s", protect_enabled)
    logger.info("Assess: %s", assess_enabled)

    if protect_enabled and settings.config.get("assess.enable"):
        logger.warning("Protect is running but Assess is enabled in local config")
        logger.warning("Defaulting to Protect behavior only")

    if (
        settings.config.get("protect.enable")
        and not settings._is_defend_enabled_in_server_features()
    ):
        logger.warning("Protect enabled in local config but disabled by Teamserver")

    if (
        settings.is_agent_config_enabled()
        and not protect_enabled
        and not assess_enabled
    ):
        logger.warning("Neither Protect nor Assess is running")


def _successful_startup_msgs():
    if not _send_startup_msg_with_retries(teamserver_messages.ServerStartup()):
        return False
    if not _send_startup_msg_with_retries(teamserver_messages.ApplicationStartup()):
        return False
    return True


def _send_startup_msg_with_retries(msg):
    attempt = 1
    while True:
        if _send_startup_msg(msg):
            return True

        attempt += 1
        if attempt > MAX_ATTEMPTS:
            break

        logger.debug("%s: will retry - sleeping for 1 second", msg.class_name)
        timer.sleep(1)

    # TODO: PYT-2612 use the commented-out logic here instead
    # logger.error(
    #     "%s: Unexpected error from Contrast UI. Will not initialize Contrast Agent."
    # )
    # return False
    return True


def _send_startup_msg(msg):
    """
    Send a single startup message (ServerStartup or ApplicationStartup) immediately.
    This function runs in the main thread, and is blocking.

    Return False if the message needs to be retried, True otherwise.
    Note that we only need to retry if we get an ambiguous response (5xx).
    """
    response = module.reporting_client.send_message(msg)
    if response is None:
        return False
    msg.process_response(response, module.reporting_client)
    return response.status_code < 500


def initialize_libraries():
    """
    If enabled, read libraries from the application
    :return: True
    """
    if not module.settings.is_analyze_libs_enabled():
        return

    # Passing callback to send message due to a circular import issue, and a
    # deadlock occurring if something is imported inside of a function running in
    # this thread
    module.library_reader = LibraryReader(
        module.settings,
        send_ts_message_func=module.reporting_client.add_message,
    )
    module.library_reader.start_library_analysis_thread()


@fail_loudly("Unable to initialize Contrast Agent Settings.", return_value=False)
def initialize_settings():
    """
    Initialize agent settings.

    Returns True on settings being initialized and False if any failure.
    """
    module.settings = Settings(
        app_name=module.app_name,
        framework_name=module.framework,
    )
    return True


def is_initialized() -> bool:
    return module.is_initialized


def get_settings():
    return module.settings


def get_reporting_client():
    return module.reporting_client


def get_routes() -> Dict[str, Route]:
    return module.routes


def update_routes(new_routes: Dict[str, Route]):
    module.routes.update(new_routes)


def is_first_request():
    return module.first_request


def set_first_request(val: bool):
    module.first_request = val


@fail_loudly("Failed to register detected framework")
def set_detected_framework(name: str):
    """
    Registers the name of the detected framework

    It is safe to call this method prior to initialization of agent state. In
    order to be effective, this method should be called prior to middleware
    initialization.
    """
    module.framework = name


@fail_loudly("Failed to register application name")
def set_application_name(name: Optional[str]):
    """
    Registers the name of the application

    It is safe to call this method prior to initialization of agent state. In
    order to be effective, this method should be called prior to middleware
    initialization.
    """
    module.app_name = name


def _log_and_warn(msg):
    warnings.warn(msg, ContrastDeprecationWarning)
    logger.warning(msg)


def _check_middleware_warnings():
    # NOTE: currently, we are skipping framework-specific middleware deprecation
    # warnings. In the future, we might want to actually deprecate these. For now,
    # however, we are holding off.
    if False and module.deprecated_middleware is not None:
        (  # pylint: disable=unpacking-non-sequence
            framework,
            is_asgi,
        ) = module.deprecated_middleware
        message = NO_SPECIFIC_MIDDLEWARE_WARNING.format(
            framework, "asgi" if is_asgi else "wsgi"
        )
        _log_and_warn(message)

    # This configuration suppresses both warnings
    if not module.settings.config.enable_automatic_middleware:
        return

    # Warn for the case where the runner is not used at all and middleware is
    # explicitly configured
    if not is_runner_in_use():
        _log_and_warn(MIDDLEWARE_WITHOUT_RUNNER_WARNING)
    # Warn for the case where middleware is explicitly configured but also detected automatically
    # An actual warning doesn't seem necessary in this case, just a log message
    elif not module.automatic_middleware.get():
        logger.warning(NO_EXPLICIT_MIDDLEWARE_WARNING)


@fail_loudly("Failed to process logs from rewriter")
def _process_rewriter_logs(logger):
    if is_runner_in_use():
        logger.debug("---- Beginning of deferred rewriter logs ----")
        process_rewriter_logs(logger)
        logger.debug("---- End of deferred rewriter logs ----")
    else:
        logger.debug("Not checking for deferred rewriter logs")

    logger.debug("Setting rewriter logger to agent logger")
    set_rewriter_logger(logger)


def initialize(name: Optional[str] = None):
    """
    If this method is called more than once per process, we use the is_initialized
    flag to only run the following work once:
        - library analysis thread initialization
        - turning on patches
        - hardcoded rule providers
        - scanning config rules
        - environment logging
        - loading common configuration
        - initializing settings

    In order to avoid a warning message, callers should check is_initialized()
    prior to calling this function.
    """
    with module.init_lock:
        if module.is_initialized:
            _check_middleware_warnings()
            logger.warning("Attempted to initialize agent state more than once")
            return

        module.id = id(module)
        name = name or __name__

        logger.info('Initializing Contrast Agent "%s" [id=%s]', name, module.id)
        logger.info("Contrast Python Agent Version: %s\n", __version__)

        if not initialize_settings():
            return

        if module.settings is None or not module.settings.is_agent_config_enabled():
            logger.warning("Contrast Agent is not enabled.")
            return

        setup_agent_logger(module.settings.config)
        setup_security_logger(module.settings.config)
        module.settings.config.log_config()

        _log_environment(module.settings)

        # NOTE: This assumes that initialize is only ever called from middleware.
        # If at some point this no longer holds, it may be necessary to pass a
        # flag to initialize indicating whether these warnings should be
        # processed or not.
        _check_middleware_warnings()

        # The rewriter is applied long before we have any settings or logger so
        # the deferred logs are finally processed here
        _process_rewriter_logs(logger)

        cs_str.init_contrast_scope_cvars()

        module.reporting_client = ReportingClient()
        module.reporting_client.start()

        if not _successful_startup_msgs():
            logger.error("Unable to initialize Contrast Agent.")
            module.settings.log_effective_config()
            return

        _warn_for_misleading_config(module.settings)

        module.settings.log_effective_config()

        # This MUST happen after the initialization calls for TeamServer messaging
        # (sending server start and application start) to ensure that TeamServer will
        # accept the messages sent by our background reporting threads
        # Skip telemetry since it is enabled later in this method
        thread_watcher.ensure_running(module, skip_telemetry=True)

        initialize_libraries()
        patch_controller.enable_patches()

        if module.settings.is_assess_enabled():
            # For now agent runtime starts before config scanning
            # this will be reset when time_limit_threshold is reached,
            # it doesn't symbolize the total agent runtime for all time.
            module.settings.agent_runtime_window = timer.now_ms()

            enable_providers()
            # NOTE: This should stay in middleware
            # scan_configs_in_thread()

        if module.settings.is_protect_enabled() and not agent_lib.initialize():
            logger.error("Fatal: Unable to initialize agent-lib. Disabling Contrast.")
            module.settings.config.put("enable", False)
            return

        logger.info(
            "Finished Initializing Contrast Agent %s [id=%s] \n\n%s\n",
            name,
            module.id,
            LOGS_SEPARATOR,
        )

        if Telemetry is not None:
            contrast.TELEMETRY = Telemetry()
            contrast.TELEMETRY.start()
            atexit.register(stop_telemetry, contrast.TELEMETRY)

        module.is_initialized = True


def detected_deprecated_middleware(*, framework: str = "", is_asgi: bool = False):
    module.deprecated_middleware = (framework, is_asgi)


@contextmanager
def automatic_middleware():
    """
    Context manager to be used by automatic middleware hooks when initializing middleware

    This enables us to determine when middleware is initialized from within our
    automatic middleware hooks without having to change the API of any
    middleware classes.
    """
    try:
        module.automatic_middleware.set(True)
        yield
    finally:
        module.automatic_middleware.set(False)


def in_automatic_middleware() -> bool:
    return module.automatic_middleware.get()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.sql_injection.default_sql_scanner import (
    DefaultSqlScanner,
)


class SqliteSqlScanner(DefaultSqlScanner):
    """
    Scanner for Sqlite
    """

    pass



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

from urllib.parse import unquote

from contrast.agent.agent_lib import does_file_path_bypass_security
from contrast.utils.decorators import cached_property
from contrast.utils.stack_trace_utils import in_custom_code

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


CUSTOM_CODE_ACCESSING_SYSTEM_FILES_REP_CODE = 0
CUSTOM_CODE_CONFIG_KEY = "detect_custom_code_accessing_system_files"

COMMON_FILE_EXPLOITS_REP_CODE = 1
COMMON_FILE_EXPLOITS_KEY = "detect_common_file_exploits"

KNOWN_SECURITY_BYPASS_MARKERS = ["::$DATA", "::$Index", "" + "\x00"]
KNOWN_SYSTEM_FILES = [
    "/proc/self",
    "etc/passwd",
    "etc/shadow",
    "etc/hosts",
    "etc/groups",
    "etc/gshadow",
    "ntuser.dat",
    "/Windows/win.ini",
    "/windows/system32/",
    "/windows/repair/",
]


class PathTraversalSinkFeatures(object):
    def check_sink_features(self, attack_vector, attack):
        is_custom_code = self._check_custom_code_accessing_system_files(attack_vector)
        new_sample = is_custom_code or self._check_file_security_bypass(
            attack_vector, is_custom_code
        )

        if new_sample:
            if attack is None:
                attack = self.build_base_attack()

            attack.add_sample(new_sample)
            attack.set_response(self.response_from_mode(self.mode))
            self.log_rule_matched(None, attack.response)

        return attack

    def _check_custom_code_accessing_system_files(self, attack_vector):
        if (
            self._is_custom_code_access_sysfile_enabled
            and PathTraversalSinkFeatures._is_custom_code_accessing_file_system(
                attack_vector
            )
        ):
            logger.debug(
                "Found custom code trying to access system file: %s", attack_vector
            )
            return self._create_path_traversal_sample(
                attack_vector,
                CUSTOM_CODE_ACCESSING_SYSTEM_FILES_REP_CODE,
            )

        return None

    def _check_file_security_bypass(self, attack_vector, is_custom_code):
        file_bypass_security = does_file_path_bypass_security(
            attack_vector, is_custom_code
        )

        if (
            self._is_common_file_exploits_enabled
            and PathTraversalSinkFeatures._contains_known_attack_signatures(
                attack_vector
            )
            and file_bypass_security
        ):
            logger.debug(
                "Found attempt to bypass file security checks: %s", attack_vector
            )
            return self._create_path_traversal_sample(
                attack_vector,
                COMMON_FILE_EXPLOITS_REP_CODE,
            )

        return None

    def _create_path_traversal_sample(self, attack_vector, rep_code):
        sample = self.build_sample(None, attack_vector)
        sample.details["path"] = attack_vector
        sample.details["path_traversal_semantic"] = {}
        sample.details["path_traversal_semantic"]["path"] = attack_vector
        sample.details["path_traversal_semantic"]["findings"] = [rep_code]
        return sample

    @cached_property
    def _is_custom_code_access_sysfile_enabled(self):
        return self.settings.is_rep_feature_enabled_for_rule(
            self.name, CUSTOM_CODE_CONFIG_KEY
        )

    @cached_property
    def _is_common_file_exploits_enabled(self):
        return self.settings.is_rep_feature_enabled_for_rule(
            self.name, COMMON_FILE_EXPLOITS_KEY
        )

    @staticmethod
    def _is_custom_code_accessing_file_system(path):
        return PathTraversalSinkFeatures._is_system_file(path) and in_custom_code()

    @staticmethod
    def _contains_known_attack_signatures(path):
        unescaped = unquoted = unquote(path)

        try:
            unescaped = unquoted.encode("utf-8").decode("unicode-escape")
        except LookupError as e:
            logger.debug(
                f"Failed to unescape attack vector path {path} with the error {e}"
            )
            return False

        try:
            realpath = os.path.realpath(unescaped).lower().rstrip("/")
        except ValueError as e:
            return str(e) == "embedded null byte"
        except TypeError as e:
            return (
                "NUL" in str(e)
                or "null byte" in str(e)
                or str(e) == "embedded NUL character"
            )
        except Exception as e:
            return "null byte" in str(e).lower()

        return any(
            [
                bypass_markers.lower().rstrip("/") in realpath
                for bypass_markers in KNOWN_SECURITY_BYPASS_MARKERS
            ]
        )

    @staticmethod
    def _is_system_file(path):
        if not path:
            return False
        unquoted = unquote(path)
        realpath = os.path.realpath(unquoted).lower().rstrip("/")

        return any(
            [
                sys_file.lower().rstrip("/") in realpath
                for sys_file in KNOWN_SYSTEM_FILES
            ]
        )



#!/usr/bin/env python

# Copyright (c) 2009, Mario Vilas
# Rob Ruana 2010
# Gil Dabah
# All rights reserved.
# Licensed under BSD.
#

__revision__ = "$Id: setup.py 603 2010-01-31 00:11:05Z qvasimodo $"

import os
import platform
import string
import shutil
import sys

from glob import glob

from distutils import log
from distutils.command.build import build
from distutils.command.build_clib import build_clib
from distutils.command.clean import clean
from distutils.command.install_lib import install_lib
from distutils.command.sdist import sdist
from distutils.core import setup, Extension
from distutils.errors import DistutilsSetupError

from shutil import ignore_patterns

def get_sources():
    """Returns a list of C source files that should be compiled to 
    create the libdistorm3 library.
    """

    return glob('src/*.c')


class custom_build(build):
    """Customized build command"""
    def run(self):
        log.info('running custom_build')
        build.run(self)


class custom_build_clib(build_clib):
    """Customized build_clib command

    This custom_build_clib will create dynamically linked libraries rather 
    than statically linked libraries.  In addition, it places the compiled 
    libraries alongside the python packages, to facilitate the use of ctypes. 
    """

    def finalize_options (self):
        # We want build-clib to default to build-lib as defined by the 
        # "build" command.  This is so the compiled library will be put 
        # in the right place along side the python code.
        self.set_undefined_options('build',
                                   ('build_lib', 'build_clib'),
                                   ('build_temp', 'build_temp'),
                                   ('compiler', 'compiler'),
                                   ('debug', 'debug'),
                                   ('force', 'force'))

        self.libraries = self.distribution.libraries
        if self.libraries: # In Python 3.0 they have a bug in check_library_list, comment it out then.
            self.check_library_list(self.libraries)

        if self.include_dirs is None:
            self.include_dirs = self.distribution.include_dirs or []
        if type(self.include_dirs) in (bytes, str):
            self.include_dirs = string.split(self.include_dirs,
                                             os.pathsep)

    def get_source_files_for_lib(self, lib_name, build_info):
        sources = build_info.get('sources', [])
        if hasattr(sources, '__call__'):
            sources = sources()
        if (sources is None or
            type(sources) not in (list, tuple) or
            len(sources) == 0):
            raise DistutilsSetupError ("in 'libraries' option (library '%s'), 'sources' must be present and must be a list of source filenames") % lib_name
        return sources

    def get_source_files(self):
        self.check_library_list(self.libraries)
        filenames = []
        for (lib_name, build_info) in self.libraries:
            sources = self.get_source_files_for_lib(lib_name, build_info)
            filenames.extend(sources)
        return filenames

    def run(self):
        log.info('running custom_build_clib')
        build_clib.run(self)

    def build_libraries (self, libraries):
        for (lib_name, build_info) in libraries:
            sources = self.get_source_files_for_lib(lib_name, build_info)
            sources = list(sources)

            log.info("building '%s' library", lib_name)

            # First, compile the source code to object files in the 
            # library directory.
            macros = build_info.get('macros')
            include_dirs = build_info.get('include_dirs')
            objects = self.compiler.compile(sources,
                output_dir=self.build_temp,
                macros=macros,
                include_dirs=include_dirs,
                extra_postargs=build_info.get('extra_compile_args', []),
                debug=self.debug)

            # Then link the object files and put the result in the 
            # package build directory.
            package = build_info.get('package', '')
            self.compiler.link_shared_lib(
                objects, lib_name,
                output_dir=os.path.join(self.build_clib, package),
                extra_postargs=build_info.get('extra_link_args', []),
                debug=self.debug,)


class custom_clean(clean):
    """Customized clean command

    Customized clean command removes .pyc files from the project, 
    as well as build and dist directories."""
    def run(self):
        log.info('running custom_clean')
        # Remove .pyc files
        if hasattr(os, 'walk'):
            for root, dirs, files in os.walk('.'):
                for f in files:
                    if f.endswith('.pyc'):
                        log.info("removing '%s'" % f)
                        try:
                            os.unlink(f)
                        except:
                            pass

        # Remove generated directories
        for dir in ['build', 'dist']:
            if os.path.exists(dir):
                log.info("removing '%s' (and everything under it)"%dir)
                try:
                    shutil.rmtree(dir, ignore_errors=True)
                except:
                    pass

        clean.run(self)


class custom_sdist(sdist):
    """Customized sdist command"""
    def run(self):
        log.info('running custom_sdist')
        sdist.run(self)


def main():
    # Just in case we are being called from a different directory
    cwd = os.path.dirname(__file__)
    if cwd:
        os.chdir(cwd)
    
    # Get the target platform
    system  = platform.system().lower()
    
    # Setup the extension module
    # Setup the library
    ext_modules = None
    libraries = None
    if 'windows' in system:
        libraries = [(
            'distorm3', dict(
            package='distorm3',
            sources=get_sources,
            include_dirs=['src', 'include'],
            extra_compile_args=['/Ox', '/Ob1', '/Oy', '"/D WIN32"',
                                '"/D DISTORM_DYNAMIC"', '"/D SUPPORT_64BIT_OFFSET"',
                                '"/D _MBCS"', '/GF', '/Gm', '/Zi', '/EHsc',
                                '/MT', '/Gy', '/W4', '/nologo', '/c', '/TC',
                                '/Fdbuild\\vc90.pdb'],
            extra_link_args=['/MANIFEST', '/DLL']))]
    elif 'darwin' in system or 'macosx' in system:
        libraries = [(
            'distorm3', dict(
            package='distorm3',
            sources=get_sources,
            include_dirs=['src', 'include'],
            extra_compile_args=['-arch', 'i386', '-arch', 'x86_64', '-O2', 
                                '-Wall', '-fPIC', '-DSUPPORT_64BIT_OFFSET', 
                                '-DDISTORM_DYNAMIC']))]
    elif 'cygwin' in system:
        libraries = [(
            'distorm3', dict(
            package='distorm3',
            sources=get_sources,
            include_dirs=['src', 'include'],
            extra_compile_args=['-fPIC', '-O2', '-Wall', 
                                '-DSUPPORT_64BIT_OFFSET', 
                                '-DDISTORM_STATIC']))]
    else:
        libraries = [(
            'distorm3', dict(
            package='distorm3',
            sources=get_sources,
            include_dirs=['src', 'include'],
            extra_compile_args=['-fPIC', '-O2', '-Wall', 
                                '-DSUPPORT_64BIT_OFFSET', 
                                '-DDISTORM_STATIC']))]
    
    options = {

    # Setup instructions
    'requires'          : ['ctypes'],
    'provides'          : ['distorm3'],
    'packages'          : ['distorm3'],
    'package_dir'       : { '' : 'python' },
    'cmdclass'          : { 'build' : custom_build,
                            'build_clib' : custom_build_clib,
                            'clean' : custom_clean, 
                            'sdist' : custom_sdist },
    'libraries'         : libraries,

    # Metadata
    'name'              : 'distorm3',
    'version'           : '3.3.4',
    'description'       : 'The goal of diStorm3 is to decode x86/AMD64' \
                          ' binary streams and return a structure that' \
                          ' describes each instruction.',
    'long_description'  : (
                        'Powerful Disassembler Library For AMD64\n'
                        'by Gil Dabah (distorm@gmail.com)\n'
                        '\n'
                        'Python bindings by Mario Vilas (mvilas@gmail.com)'
                        ),
    'author'            : 'Gil Dabah',
    'author_email'      : 'distorm'+chr(64)+'gmail'+chr(0x2e)+'com',
    'maintainer'        : 'Gil Dabah',
    'maintainer_email'  : 'distorm'+chr(64)+'gmail'+chr(0x2e)+'com',
    'url'               : 'https://github.com/gdabah/distorm/',
    'download_url'      : 'https://github.com/gdabah/distorm/',
    'platforms'         : ['cygwin', 'win', 'linux', 'macosx'],
    'classifiers'       : [
                        'License :: OSI Approved :: BSD License',
                        'Development Status :: 5 - Production/Stable',
                        'Intended Audience :: Developers',
                        'Natural Language :: English',
                        'Operating System :: Microsoft :: Windows',
                        'Operating System :: MacOS :: MacOS X',
                        'Operating System :: POSIX :: Linux',
                        'Programming Language :: Python :: 2.7',
                        'Programming Language :: Python :: 3.5',
                        'Topic :: Software Development :: Disassemblers',
                        'Topic :: Software Development :: Libraries :: Python Modules',
                        ]
    }

    # Call the setup function
    setup(**options)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
# pylint: disable=too-many-lines
import os
import pathlib
from datetime import datetime, timezone
import json
from urllib.parse import urlparse

from contrast import AGENT_CURR_WORKING_DIR

from contrast.utils.decorators import fail_loudly
from contrast.agent.exclusions import Exclusions, TsExclusion, ExclusionType
from contrast.agent.framework import Framework, Server
from contrast.agent.protect.mixins.REP_settings import SettingsREPMixin
from contrast.agent.reaction_processor import ReactionProcessor
from contrast.agent.protect.rule import ProtectionRule
from contrast.configuration import (
    AgentConfig,
    SamplingLockConfig,
    ConfigOptionSrc,
    CONTRAST_UI_SRC,
    USER_CONFIGURATION_FILE_SRC,
    DEFAULT_VALUE_SRC,
    ASSESS_DISABLED_RULE_CONFIG_KEY,
)
from contrast.utils.decorators import cached_property
from contrast.utils.loggers.logger import reset_agent_logger
from contrast.utils.singleton import Singleton
from contrast.utils.string_utils import truncate
from contrast.utils.timer import now_ms
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

ASSESS_STACKTRACES = "assess.stacktraces"
EXPORTED_CONFIG_FILE_NAME = "contrast_connection.json"


class ServerFeatures:
    def __init__(self, features):
        self.features = features or {}

    @cached_property
    def log_file(self):
        return self.features.get("features", {}).get("logFile", "")

    @cached_property
    def log_level(self):
        return self.features.get("features", {}).get("logLevel", "")

    @cached_property
    def assess_enabled(self):
        return (
            self.features.get("features", {})
            .get("assessment", {})
            .get("enabled", False)
        )

    @cached_property
    def protect_enabled(self):
        return self.features.get("features", {}).get("defend", {}).get("enabled", False)


class ApplicationSettings:
    def __init__(self, settings_json=None):
        self._raw_settings = settings_json or {}

    @cached_property
    def disabled_assess_rules(self):
        return (
            self._raw_settings.get("settings", {})
            .get("assessment", {})
            .get("disabledRules", [])
        )

    @cached_property
    def protection_rules(self):
        return [
            ProtectRule(r)
            for r in self._raw_settings.get("settings", {})
            .get("defend", {})
            .get("protectionRules", [])
        ]

    @cached_property
    def session_id(self):
        return (
            self._raw_settings.get("settings", {})
            .get("assessment", {})
            .get("session_id", None)
        )

    @cached_property
    def sensitive_data_masking_policy(self):
        return self._raw_settings.get("settings", {}).get(
            "sensitive_data_masking_policy"
        )

    @cached_property
    def exclusions(self):
        """
        Build and return a plain list of TsExclusions based on raw application settings.
        TsExclusions represent the exclusions more or less exactly as they were given to
        us by teamserver.

        The list returned by this property will be converted into our own Exclusions
        class later, which is more useful when processing exclusion rules.
        """
        ts_exclusions = self._raw_settings.get("settings", {}).get("exceptions")
        if ts_exclusions is None:
            return None

        exclusions = []

        for exclusion in ts_exclusions.get("urlExceptions", []):
            exclusions.append(
                TsExclusion(
                    ExclusionType.URL_EXCLUSION_TYPE,
                    exclusion.get("name", ""),
                    exclusion.get("modes", []),
                    exclusion.get("matchStrategy", ""),
                    exclusion.get("protectionRules", []),
                    exclusion.get("assessmentRules", []),
                    exclusion.get("urls", []),
                )
            )

        for exclusion in ts_exclusions.get("inputExceptions", []):
            exclusions.append(
                TsExclusion(
                    ExclusionType.INPUT_EXCLUSION_TYPE,
                    exclusion.get("name", ""),
                    exclusion.get("modes", []),
                    exclusion.get("matchStrategy", ""),
                    exclusion.get("protectionRules", []),
                    exclusion.get("assessmentRules", []),
                    exclusion.get("urls", []),
                    exclusion.get("inputType", ""),
                    exclusion.get("inputName", ""),
                )
            )

        return exclusions

    # note: there are more fields on ApplicationSettings that we currently don't use


class ProtectRule:
    def __init__(self, protect_rule_json):
        self._raw_rule = protect_rule_json or {}

    @cached_property
    def id(self):
        return self._raw_rule.get("id", "")

    @cached_property
    def mode(self):
        raw_mode = self._raw_rule.get("mode", "OFF")
        raw_block_at_entry = self._raw_rule.get("blockAtEntry", False)
        return {
            "MONITORING": ProtectionRule.MONITOR,
            "BLOCKING": (
                ProtectionRule.BLOCK_AT_PERIMETER
                if raw_block_at_entry
                else ProtectionRule.BLOCK
            ),
            "OFF": ProtectionRule.OFF,
        }[raw_mode]


class Settings(Singleton, SettingsREPMixin):
    def init(self, app_name=None, framework_name=None):
        """
        Agent settings for the entire lifetime of the agent.

        Singletons should override init, not __init__.
        """
        self.config = None
        self.config_features = {}
        self.exclusions = None
        self.last_server_update_time_ms = 0
        self.last_app_update_time_ms = 0
        self.heartbeat = None
        self.server_settings_poll = None
        self.framework = Framework(framework_name)
        self.server = Server()
        self.sys_module_count = 0

        # Server
        self.server_name = None
        self.server_path = None
        self.server_type = None

        self.exclusion_matchers = []

        # Rules
        self.protect_rules = dict()

        # circular import
        from contrast.agent.assess.rules.response.autocomplete_missing_rule import (
            AutocompleteMissingRule,
        )
        from contrast.agent.assess.rules.response.cache_controls_rule import (
            CacheControlsRule,
        )
        from contrast.agent.assess.rules.response.clickjacking_rule import (
            ClickjackingRule,
        )

        from contrast.agent.assess.rules.response.x_content_type_rule import (
            XContentTypeRule,
        )

        from contrast.agent.assess.rules.response.csp_header_missing_rule import (
            CspHeaderMissingRule,
        )

        from contrast.agent.assess.rules.response.x_xss_protection_disabled_rule import (
            XXssProtectionDisabledRule,
        )

        from contrast.agent.assess.rules.response.hsts_header_rule import HstsHeaderRule

        from contrast.agent.assess.rules.response.csp_header_insecure_rule import (
            CspHeaderInsecureRule,
        )

        from contrast.agent.assess.rules.response.parameter_pollution_rule import (
            ParameterPollutionRule,
        )

        self.assess_response_rules = [
            AutocompleteMissingRule(),
            CacheControlsRule(),
            ClickjackingRule(),
            XContentTypeRule(),
            CspHeaderMissingRule(),
            XXssProtectionDisabledRule(),
            HstsHeaderRule(),
            CspHeaderInsecureRule(),
            ParameterPollutionRule(),
        ]

        # Initialize config
        self.config = AgentConfig()

        self.direct_teamserver_server_features = None
        self.direct_ts_app_settings = ApplicationSettings()

        self.disabled_assess_rules = set(
            self.config.get(ASSESS_DISABLED_RULE_CONFIG_KEY, [])
        )

        # Initialize application metadata
        self.app_name = self.get_app_name(app_name)

        self.agent_runtime_window = now_ms()

        logger.info("Contrast Agent finished loading settings.")

    def _is_defend_enabled_in_server_features(self):
        return bool(
            self.direct_teamserver_server_features
            and self.direct_teamserver_server_features.protect_enabled
        )

    @cached_property
    def is_proxy_enabled(self):
        return self.config.get("api.proxy.enable")

    @cached_property
    def proxy_url(self):
        return self.config.get("api.proxy.url")

    @cached_property
    def proxy_scheme(self):
        scheme = urlparse(self.proxy_url).scheme or "http"
        return scheme

    @cached_property
    def is_cert_verification_enabled(self):
        return self.config.get("api.certificate.enable")

    @cached_property
    def ignore_cert_errors(self):
        return self.config.get("api.certificate.ignore_cert_errors")

    @cached_property
    def ca_file(self):
        return self.config.get("api.certificate.ca_file")

    @cached_property
    def client_cert_file(self):
        return self.config.get("api.certificate.cert_file")

    @cached_property
    def client_private_key(self):
        return self.config.get("api.certificate.key_file")

    @cached_property
    def api_service_key(self):
        return self.config.get("api.service_key")

    @cached_property
    def api_url(self):
        """Normalizes the URL to remove any whitespace or trailing slash"""
        return self.config.get("api.url").strip().rstrip("/")

    @cached_property
    def api_key(self):
        return self.config.get("api.api_key")

    @cached_property
    def api_user_name(self):
        return self.config.get("api.user_name")

    def is_agent_config_enabled(self):
        if self.config is None:
            return True

        return self.config.get("enable", True)

    @cached_property
    def is_rewriter_enabled(self):
        return self.config.get("agent.python.rewrite")

    @cached_property
    def is_policy_rewriter_enabled(self):
        return self.config.should_apply_policy_rewrites

    @cached_property
    def is_profiler_enabled(self):
        return self.config.get("agent.python.enable_profiler")

    @cached_property
    def max_sources(self):
        return self.config.get("assess.max_context_source_events")

    @cached_property
    def max_propagation(self):
        return self.config.get("assess.max_propagation_events")

    @cached_property
    def max_vulnerability_count(self):
        """Max number of vulnerabilities per rule type to report for one
        agent run `time_limit_threshold` time period"""
        return self.config.get("assess.max_rule_reported")

    @cached_property
    def agent_runtime_threshold(self):
        return self.config.get("assess.time_limit_threshold")

    @cached_property
    def app_path(self):
        return self.config.get("application.path")

    @cached_property
    def app_version(self):
        return self.config.get("application.version")

    @property
    def pid(self):
        """
        pid is used in our CMDi protect rule.

        pid must be unique for each worker process of an app.
        :return: int current process id
        """
        return os.getpid()

    def get_app_name(self, app_name):
        if self.config.get("application.name"):
            return self.config.get("application.name")

        return app_name if app_name else "root"

    def establish_heartbeat(self):
        """
        Initialize Heartbeat between Agent and TS if it has not been already initialized.
        """
        if self.heartbeat is None:
            # Circular import
            from contrast.agent.heartbeat_thread import HeartbeatThread

            self.heartbeat = HeartbeatThread()
            self.heartbeat.start()

    def establish_server_settings_poll(self):
        """
        Initialize Server Settings poll between Agent and TS if it has not been already initialized.
        """
        if self.server_settings_poll is None:
            # Circular import
            from contrast.agent.server_settings_poll import ServerSettingsPoll

            self.server_settings_poll = ServerSettingsPoll()
            self.server_settings_poll.start()

    def set_protect_assess_config_source(self):
        canonical_name_assess = "assess.enable"
        canonical_name_protect = "protect.enable"

        if self.config is None or self.direct_teamserver_server_features is None:
            return

        if (
            self.config.get(canonical_name_assess, None) is None
            or self.config.get_src(canonical_name_assess).source == CONTRAST_UI_SRC
        ):
            self.config.put(
                canonical_name_assess,
                self.direct_teamserver_server_features.assess_enabled,
                src=ConfigOptionSrc(
                    name="contrast.assess.enable", source=CONTRAST_UI_SRC
                ),
            )

        if (
            self.config.get(canonical_name_protect, None) is None
            or self.config.get_src(canonical_name_protect).source == CONTRAST_UI_SRC
        ):
            self.config.put(
                canonical_name_protect,
                self.direct_teamserver_server_features.protect_enabled,
                src=ConfigOptionSrc(
                    name="contrast.protect.enable", source=CONTRAST_UI_SRC
                ),
            )

    def apply_ts_feature_settings(self, response_body):
        self.direct_teamserver_server_features = ServerFeatures(response_body)

        if (
            self.direct_teamserver_server_features.features.get("features", {})
            .get("assessment", {})
            .get("sampling")
        ):
            self.update_sampling()

        self.update_logger_from_features()

        self.log_server_features(self.direct_teamserver_server_features)

        self.set_protect_assess_config_source()

        self.set_protect_rules()

        self.last_server_update_time_ms = now_ms()

    def process_ts_reactions(self, response_body):
        # App startup/activity wrap reactions in a settings dict whereas
        # Server startup/activity has it at the top level response dict
        reactions = response_body.get("settings", {}).get("reactions", None)

        if not reactions:
            reactions = response_body.get("reactions", None)

        if not reactions:
            return

        ReactionProcessor.process(reactions, self)

    def log_server_features(self, server_features):
        """
        Record server features received from teamserver
        """
        logger.debug(
            "Received updated server features logFile=%s logLevel=%s Protect=%s Assess=%s Sampling=%s",
            server_features.log_file,
            server_features.log_level,
            server_features.protect_enabled,
            server_features.assess_enabled,
            server_features.features.get("features", {})
            .get("assessment", {})
            .get("sampling", None),
        )

    @property
    def code_exclusion_matchers(self):
        return [x for x in self.exclusion_matchers if x.is_code]

    def load_exclusions_from_server(self, exclusions):
        if exclusions:
            self.exclusions = Exclusions(exclusions)

    def evaluate_exclusions(self, context, path):
        if self.exclusions:
            if self.exclusions.evaluate_url_exclusions(context, path):
                # Stop analyzing this endpoint since the URL exclusion applies
                return True

            if self.exclusions.input_exclusions:
                self.exclusions.set_input_exclusions_by_url(context, path)

        return False

    def apply_ts_app_settings(self, response_json):
        self.direct_ts_app_settings = ApplicationSettings(response_json)
        if self.direct_ts_app_settings.exclusions:
            self.load_exclusions_from_server(self.direct_ts_app_settings.exclusions)

        # This is the only place session_id is set by TS.
        # If session id is set in the config, that value will be echoed back by TS
        if self.direct_ts_app_settings.session_id:
            self.config.session_id = self.direct_ts_app_settings.session_id

    def update_logger_from_features(self):
        if self.direct_teamserver_server_features is not None:
            logger_reset = reset_agent_logger(
                self.direct_teamserver_server_features.log_file,
                self.direct_teamserver_server_features.log_level,
                self.config,
            )
        else:
            logger_reset = False

        if logger_reset and self.config is not None:
            self.config.log_config()

    def is_inventory_enabled(self):
        """
        inventory.enable = false: Disables both route coverage and library analysis and reporting
        """
        return self.config.get("inventory.enable", True)

    def is_analyze_libs_enabled(self):
        """
        inventory.analyze_libraries = false: Disables only library analysis/reporting
        """
        return (
            self.config is not None
            and self.config.get("inventory.analyze_libraries", True)
            and self.is_inventory_enabled()
        )

    def is_assess_enabled(self):
        """
        We do not allow assess and defend to be on at the same time. As defend
        is arguably the more important of the two, it will take precedence

        The agent config may enable assess even if it is turned off in TS. This
        allows unlicensed apps to send findings to TS, where they will appear
        as obfuscated results.
        """

        # https://contrast.atlassian.net/browse/PROD-530
        if self.config is None:
            return False

        assess_enabled = self.config.get("assess.enable", None)
        if assess_enabled is not None:
            return assess_enabled

        return (
            self.direct_teamserver_server_features
            and self.direct_teamserver_server_features.assess_enabled
        )

    def is_protect_enabled(self):
        """
        Protect is enabled only if both configuration and server features enable it.
        """
        if self.config is None:
            return False
        config_protect_enabled = self.config.get("protect.enable", None)
        if config_protect_enabled is not None:
            return config_protect_enabled

        return self._is_defend_enabled_in_server_features()

    def set_protect_rules(self):
        if not self._is_defend_enabled_in_server_features():
            self.protect_rules = dict()
            return

        from contrast.agent.protect.rule.rules_builder import build_protect_rules

        self.protect_rules = build_protect_rules()

    def get_server_name(self):
        """
        Hostname of the server

        Default is socket.gethostname() or localhost
        """
        if self.server_name is None:
            self.server_name = self.config.get("server.name")

        return self.server_name

    def get_server_path(self):
        """
        Working Directory of the server

        Default is root
        """
        if self.server_path is None:
            self.server_path = self.config.get("server.path") or truncate(
                AGENT_CURR_WORKING_DIR
            )

        return self.server_path

    def get_server_type(self):
        """
        Web Framework of the Application either defined in common config or via discovery.
        """
        if self.server_type is None:
            self.server_type = (
                self.config.get("server.type") or self.framework.name_lower
            )

        return self.server_type

    @property
    def response_scanning_enabled(self):
        return self.is_assess_enabled() and self.config.get(
            "assess.enable_scan_response"
        )

    def is_assess_rule_disabled(self, rule_id):
        """
        Rules disabled in config override all disabled rules from TS per common config
        """
        app_settings = self.direct_ts_app_settings

        return (
            rule_id in self.disabled_assess_rules
            if self.disabled_assess_rules
            else app_settings and rule_id in app_settings.disabled_assess_rules
        )

    def enabled_response_rules(self):
        return [
            rule
            for rule in self.assess_response_rules
            if not self.is_assess_rule_disabled(rule.name)
        ]

    def is_collect_stacktraces_all(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "ALL"

    def is_collect_stacktraces_some(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "SOME"

    def is_collect_stacktraces_none(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "NONE"

    def build_proxy_url(self):
        if self.proxy_url:
            return {self.proxy_scheme: self.proxy_url}

        return {}

    def update_sampling(self):
        server_sampling = (
            self.direct_teamserver_server_features.features.get("features")
            .get("assessment")
            .get("sampling")
        )

        sampling_lock_config = SamplingLockConfig()

        if not sampling_lock_config.lock_enable:
            value = server_sampling.get("enabled", None)
            if value is None:
                value = server_sampling.get("enable", False)

            self.config.put(
                "assess.sampling.enable",
                value,
                src=ConfigOptionSrc(
                    name="contrast.assess.sampling.enable", source=CONTRAST_UI_SRC
                ),
            )

        if not sampling_lock_config.lock_baseline:
            self.config.put(
                "assess.sampling.baseline",
                server_sampling.get("baseline"),
                src=ConfigOptionSrc(
                    name="contrast.assess.sampling.baseline", source=CONTRAST_UI_SRC
                ),
            )

        if not sampling_lock_config.lock_frequency:
            self.config.put(
                "assess.sampling.request_frequency",
                server_sampling.get("frequency"),
                src=ConfigOptionSrc(
                    name="contrast.assess.sampling.request_frequency",
                    source=CONTRAST_UI_SRC,
                ),
            )

        if not sampling_lock_config.lock_window:
            self.config.put(
                "assess.sampling.window_ms",
                server_sampling.get("window"),
                src=ConfigOptionSrc(
                    name="contrast.assess.sampling.window_ms", source=CONTRAST_UI_SRC
                ),
            )

        logger.debug(
            "Updated sampling setting with TeamServer sampling settings: %s",
            server_sampling,
        )

    def _export_agent_config(self, redacted_config, effective_config):
        """
        The purpose of this function is to export as many of the configuration values found in AgentConfig
        to the effective_config list. This list will be logged to disk. Some values that where set using
        the Contrast UI where saved to AgentConfig in attempt to have a single place storing all settings.

        In some cases this wasn't immediately feasible due to when and how we reload those settings
        """
        # List of configs that have their own custom export function due to how they are configured between
        # Contrast UI and AgentConfig
        custom_export_function = (
            "protect.rules.cmd-injection.mode",
            "protect.rules.method-tampering.mode",
            "protect.rules.nosql-injection.mode",
            "protect.rules.path-traversal.mode",
            "protect.rules.reflected-xss.mode",
            "protect.rules.sql-injection.mode",
            "protect.rules.ssrf.mode",
            "protect.rules.unsafe-file-upload.mode",
            "protect.rules.xxe.mode",
            ASSESS_DISABLED_RULE_CONFIG_KEY,
        )

        for key in redacted_config:
            if key in custom_export_function:
                continue

            config_option_src = self.config.get_src(key, None)

            if config_option_src is None:
                logger.debug("WARNING: source not found for the key %s", key)
                continue

            if (
                config_option_src is not None
                and config_option_src.source == CONTRAST_UI_SRC
                and self.contrast_ui_status is not None
            ):
                # The spec states this value is not included if we can't communicate with Contrast UI
                continue

            value = str(redacted_config[key])

            config_entry = {
                "canonical_name": key,
                "name": config_option_src.name if config_option_src else "",
                "source": config_option_src.source if config_option_src else "",
                "value": value,
                "filename": self.config.yaml_filename
                if config_option_src.source == USER_CONFIGURATION_FILE_SRC
                else "",
            }

            effective_config.append(config_entry)

    def _export_disabled_assess_config(self, effective_config):
        # Export assess disabled assess rules
        canonical_name = name = ASSESS_DISABLED_RULE_CONFIG_KEY
        source = DEFAULT_VALUE_SRC
        value = []

        if self.config.get(canonical_name, []):
            source = self.config.get_src(canonical_name, None)
            if source is not None:
                source = source.source
            else:
                logger.debug("WARNING: source not found for the key %s", canonical_name)
                return

            value = str(self.config.get(canonical_name, []))

        if (
            source == DEFAULT_VALUE_SRC
            and self.direct_ts_app_settings.disabled_assess_rules
        ):
            source = CONTRAST_UI_SRC
            name = f"contrast.{name}"
            value = str(self.direct_ts_app_settings.disabled_assess_rules)

        if source == CONTRAST_UI_SRC and self.contrast_ui_status is not None:
            # The spec states this value is not included if we can't communicate with Contrast UI
            return

        config_entry = {
            "canonical_name": canonical_name,
            "name": name,
            "source": source,
            "value": value,
            "filename": self.config.yaml_filename
            if source == USER_CONFIGURATION_FILE_SRC
            else "",
        }

        effective_config.append(config_entry)

    def _export_protect_rules_mode(self, effective_config):
        # Export mode for each rule that supports Mode
        for rule in self.protect_rules.values():
            canonical_name = rule.config_rule_path_mode
            source_node = rule.config_source_mode
            source = source_node.source
            name = source_node.name

            if source == CONTRAST_UI_SRC and self.contrast_ui_status is not None:
                return

            config_entry = {
                "canonical_name": canonical_name,
                "name": name,
                "source": source,
                "value": str(rule.mode),
                "filename": self.config.yaml_filename
                if source == USER_CONFIGURATION_FILE_SRC
                else "",
            }

            effective_config.append(config_entry)

    def _export_contrast_ui_config(self, effective_config):
        """
        The purpose of this function is export the remainder of the agent config settings
        that are set in the Contrast UI.
        """
        if self.direct_ts_app_settings is None or self.contrast_ui_status is not None:
            return

        name = "contrast.exclusions"
        value = str(
            self.direct_ts_app_settings._raw_settings.get("settings", {}).get(
                "exceptions"
            )
        )

        config_entry = {
            "canonical_name": name,
            "name": name,
            "source": CONTRAST_UI_SRC,
            "value": value,
            "filename": "",
        }

        effective_config.append(config_entry)

        name = "contrast.sensitive_data_masking_policy"
        value = str(self.direct_ts_app_settings.sensitive_data_masking_policy)

        config_entry = {
            "canonical_name": name,
            "name": name,
            "source": CONTRAST_UI_SRC,
            "value": value,
            "filename": "",
        }

        effective_config.append(config_entry)

    @cached_property
    def contrast_ui_status(self):
        config_name = "enable"
        error_msg = "Unable to connect to Contrast; configuration details from the Contrast UI will not be included."

        if (
            self.config.get(config_name) is False
            and self.config.get_src(config_name).source == CONTRAST_UI_SRC
        ):
            # This indicates we were told to shut down by Contrast UI
            return error_msg

        # TODO: PYT-2612 seems like _send_startup_msg_with_retries in agent_state always returns true
        return None

    @fail_loudly("Failed to export effective config")
    def log_effective_config(self):
        status = self.contrast_ui_status or self.config.config_status or "Success"
        report_creation_time = datetime.now(timezone.utc).isoformat()
        config = {
            "report_create": report_creation_time,
            "config": {
                "status": status,
                "effective_config": [],
            },
        }

        redacted_config = self.config.get_redacted_config()
        effective_config = config["config"]["effective_config"]

        self._export_agent_config(redacted_config, effective_config)
        self._export_disabled_assess_config(effective_config)
        self._export_protect_rules_mode(effective_config)
        self._export_contrast_ui_config(effective_config)

        path = pathlib.Path(self.config.get("agent.logger.path", "")).parent.resolve()
        filename = f"{path}/{EXPORTED_CONFIG_FILE_NAME}"

        with open(filename, "w") as f:
            json.dump(config, f, indent=4)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import zlib


class Digest(object):
    def __init__(self):
        self.crc32 = 0

    def finish(self):
        return str(self.crc32)

    def update(self, value):
        if not value:  # None or empty string
            return

        self.crc32 = zlib.crc32(str(value).encode(), self.crc32)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.configuration import AgentConfig
from contrast.reporting import teamserver_messages, reporting_client
from contrast.utils.loggers import logger


def validate_config():
    try:
        exit_status = _validate()
        sys.exit(exit_status)
    except Exception as e:
        _log("Unexpected error while validating agent config:")
        _log(f"{e!r}")
        _log("Unable to validate agent config")
        sys.exit(1)


def _log(msg):
    print(f"[contrast-validate-config] {msg}")


def _validate():
    """
    Validates the config and returns an appropriate exit status
    """
    _log("Validating Contrast configuration")
    _log("You may see agent logs in structured log format")

    logger.setup_basic_agent_logger()

    if not _check_config():
        return 1

    if not _check_connection():
        return 1

    return 0


def _valid_proxy(config):
    """
    If proxy use is enabled, verify the proxy url is provided.
    """
    enabled = config.get("api.proxy.enable")
    if not enabled:
        # If proxy isn't enabled then it's "valid"
        _log("Proxy use disabled")
        return True

    url = config.get("api.proxy.url")
    if url:
        # Just check that's a populated string, not that it's a url
        _log("Proxy use enabled")
        return True

    _log("Proxy enabled but no proxy url provided")
    return False


def _valid_cert(config):
    """
    If certificate use is enabled, verify that the three required files are present.
    """
    enabled = config.get("api.certificate.enable")
    if not enabled:
        # If certificate use isn't enabled then it's "valid"
        _log("Certificate use disabled")
        return True

    ca_file = config.get("api.certificate.ca_file")
    if not ca_file:
        _log("CA file path not provided")
        return False

    cert_file = config.get("api.certificate.cert_file")
    if not cert_file:
        _log("Cert file path not provided")
        return False

    key_file = config.get("api.certificate.key_file")
    if not key_file:
        _log("Key file path not provided")
        return False

    _log("Certificate use enabled")
    return True


def _check_config():
    _log("Loading config")

    config = AgentConfig()

    _log("Config loaded successfully")

    _log("Checking API configuration")

    missing_values = config.check_for_api_config()
    if missing_values:
        for missing_value in missing_values:
            _log(f"Missing required config value: {missing_value}")
        return False

    if not _valid_proxy(config) or not _valid_cert(config):
        return False

    return True


def _check_connection():
    client = reporting_client.ReportingClient()
    msg = teamserver_messages.ServerActivity()

    _log("Sending test request to Contrast UI")

    resp = client.send_message(msg)

    if resp is None:
        _log("Request failed")
        _log("Unable to establish a connection with current configuration")
        return False

    _log(f"Response: {resp.status_code} {resp.reason}")
    if resp.text:
        _log(resp.text)

    # note: status code 400 for ServerActicity just means the server doesn't exist
    # it still indicates a successful connection to teamserver
    # TODO: PYT-2080 Use ServerCreate instead

    if resp.status_code == 401:
        _log("You are connecting to the Contrast UI but have improper authorization")
        return False

    if resp.status_code > 401:
        _log("You are connecting to the Contrast UI but are seeing an unexpected error")
        return False

    _log(f"{resp.status_code} status code indicates success for this endpoint")
    _log("Connection to the Contrast UI successful")
    return True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from enum import Enum, auto

from contrast.api import TypeCheckedProperty
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class TraceEventObject(object):
    tracked = TypeCheckedProperty(bool)
    value = TypeCheckedProperty(str)
    # list of contrast.api.TraceTaintRange
    ranges = TypeCheckedProperty(list)


class TraceEventSignature(object):
    return_type = TypeCheckedProperty(str)
    class_name = TypeCheckedProperty(str)
    method_name = TypeCheckedProperty(str)
    arg_types = TypeCheckedProperty(list)
    constructor = TypeCheckedProperty(bool)
    void_method = TypeCheckedProperty(bool)
    flags = TypeCheckedProperty(int)


class ParentObjectId(object):
    id = TypeCheckedProperty(int)


class TraceEventSource(object):
    type = TypeCheckedProperty(str)
    name = TypeCheckedProperty(str)


class TraceStack(object):
    signature = TypeCheckedProperty(str)
    declaring_class = TypeCheckedProperty(str)
    method_name = TypeCheckedProperty(str)
    file_name = TypeCheckedProperty(str)
    line_number = TypeCheckedProperty(int)
    type = TypeCheckedProperty(str)
    eval = TypeCheckedProperty(str)


class TraceTaintRange(object):
    tag = TypeCheckedProperty(str)
    range = TypeCheckedProperty(str)


class TraceEvent(object):
    class EventType(Enum):
        METHOD = auto()
        PROPAGATION = auto()
        TAG = auto()

    class Action(Enum):
        CREATION = auto()
        PROPAGATION = auto()
        TRIGGER = auto()
        TAG = auto()
        A2A = auto()
        A2P = auto()
        A2O = auto()
        A2R = auto()
        O2A = auto()
        O2P = auto()
        O2O = auto()
        O2R = auto()
        P2A = auto()
        P2P = auto()
        P2O = auto()
        P2R = auto()

    action = TypeCheckedProperty(Action, constructor_arg=Action.CREATION)
    type = TypeCheckedProperty(EventType, constructor_arg=EventType.METHOD)
    timestamp_ms = TypeCheckedProperty(int)
    thread = TypeCheckedProperty(str)
    signature = TypeCheckedProperty(TraceEventSignature)
    field_name = TypeCheckedProperty(str)
    context = TypeCheckedProperty(str)
    code = TypeCheckedProperty(str)
    object = TypeCheckedProperty(TraceEventObject)
    ret = TypeCheckedProperty(TraceEventObject)
    # list of TraceEventObject
    args = TypeCheckedProperty(list)
    # list of TraceStack
    stack = TypeCheckedProperty(list)
    # list of TraceEventSource
    event_sources = TypeCheckedProperty(list)
    tags = TypeCheckedProperty(set)
    source = TypeCheckedProperty(str)
    target = TypeCheckedProperty(str)
    # list of contrast.api.TraceTaintRange
    taint_ranges = TypeCheckedProperty(list)
    object_id = TypeCheckedProperty(int)
    # list of repeated contrast.api.ParentObjectId
    parent_object_ids = TypeCheckedProperty(list)

    def to_json(self):
        return {
            "action": self.action.name,
            "args": [
                {
                    # "hash": 0,  # not required
                    "tracked": arg.tracked,
                    "value": arg.value,
                }
                for arg in self.args
            ],
            # "code": "string",  # currently unused; maybe useful in the future
            "eventSources": [
                {
                    "sourceName": s.name,
                    "sourceType": s.type,
                }
                for s in self.event_sources
            ],
            "fieldName": self.field_name,
            "object": {
                # "hash": 0,  # not required
                "tracked": self.object.tracked,
                "value": self.object.value,
            },
            "objectId": self.object_id,
            "parentObjectIds": [{"id": p.id} for p in self.parent_object_ids],
            # properties not used for dataflow rules
            # "properties": [{"key": "string", "value": "string"}],
            "ret": {
                # "hash": 0,  # not required
                "tracked": self.ret.tracked,
                "value": self.ret.value,
            },
            "signature": {
                "argTypes": list(self.signature.arg_types),
                "className": self.signature.class_name,
                "constructor": self.signature.constructor,
                # not required
                # "expressionType": "MEMBER_EXPRESSION",
                # "flags": 0,  # java only
                "methodName": self.signature.method_name,
                # not required
                # "operator": "string",
                "returnType": self.signature.return_type,
                # "signature": "string",  # deprecated
                "voidMethod": self.signature.void_method,
            },
            "source": self.source,
            "stack": [
                {
                    "eval": s.eval,
                    "file": s.file_name,
                    "lineNumber": s.line_number,
                    "method": s.method_name,
                    "signature": s.signature,
                    "type": s.type,
                }
                for s in self.stack
            ],
            # "tags": "string",  # we don't save this to the object
            "taintRanges": [
                {
                    "tag": t.tag,
                    "range": t.range,
                }
                for t in self.taint_ranges
            ],
            "target": self.target,
            # "thread": "string",  # not required
            "time": self.timestamp_ms,
            "type": self.type.name,
        }



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import argparse
import os
import sys
import shutil

from contrast import __file__
from contrast.configuration import AgentConfig
from contrast_rewriter import ENABLE_REWRITER, REWRITE_FOR_PYTEST
from contrast_vendor import structlog as logging

USING_RUNNER_ENV_VAR = "__CONTRAST_USING_RUNNER"

logger = logging.getLogger("contrast")

DESCRIPTION = """
The command-line runner for the Contrast Python Agent.
"""

USAGE = "%(prog)s [-h] -- cmd [cmd ...]"

EPILOG = """
Insert this command before the one you usually use to start your webserver
to apply Contrast's instrumentation. See our public documentation for details:
https://docs.contrastsecurity.com/en/python.html
"""


def runner() -> None:
    logger.info("Starting Contrast Agent runner pre-process")
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        usage=USAGE,
        epilog=EPILOG,
    )
    parser.add_argument(
        "--rewrite-for-pytest", action="store_true", help=argparse.SUPPRESS
    )
    # if you add public arguments here, update USAGE accordingly
    parser.add_argument("cmd", nargs="+")

    parsed = parser.parse_args()

    config = AgentConfig()

    loader_path = os.path.join(os.path.dirname(__file__), "loader")
    os.environ["PYTHONPATH"] = os.path.pathsep.join([loader_path] + sys.path)
    os.environ[USING_RUNNER_ENV_VAR] = "true"

    if parsed.rewrite_for_pytest:
        os.environ[REWRITE_FOR_PYTEST] = "true"
    elif config.should_rewrite:
        os.environ[ENABLE_REWRITER] = "true"

    os.execl(shutil.which(parsed.cmd[0]), *parsed.cmd)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import print_function
import os
import sys
from contrast.agent.assess.string_tracker import StringTracker
from contrast.utils.context_tracker import ContextTracker
from contrast.version import __version__
from contrast.assess_extensions import cs_str
from contrast.agent.assess.utils import get_properties

# process globals
CS__CONTEXT_TRACKER = ContextTracker()
STRING_TRACKER = StringTracker()
TELEMETRY = None

# PERF: These values are constant for the lifetime of the agent,
# so we compute them only once instead of potentially computing
# them hundreds of times.
AGENT_CURR_WORKING_DIR = os.getcwd()
SORTED_SYS_PATH = sorted(sys.path, key=len, reverse=True)


def telemetry_disabled() -> bool:
    return os.environ.get("CONTRAST_AGENT_TELEMETRY_OPTOUT", "").lower() in [
        "1",
        "true",
    ]


def get_canonical_version() -> str:
    return ".".join(__version__.split(".")[:3])


class SecurityException(Exception):
    """
    Exception raised by Contrast Protect to block attacks. Full attack details are
    reported to the Contrast UI.
    """

    def __init__(self, *, rule_name: str) -> None:
        super().__init__(
            f"Contrast Protect blocked an attack for rule: {rule_name}. See Contrast UI"
            " for full details."
        )



# coding: utf-8

from contrast_vendor.ruamel.yaml.emitter import Emitter
from contrast_vendor.ruamel.yaml.serializer import Serializer
from contrast_vendor.ruamel.yaml.representer import (
    Representer,
    SafeRepresenter,
    BaseRepresenter,
    RoundTripRepresenter,
)
from contrast_vendor.ruamel.yaml.resolver import Resolver, BaseResolver, VersionedResolver

from typing import Any, Dict, List, Union, Optional  # NOQA
from contrast_vendor.ruamel.yaml.compat import StreamType, VersionType  # NOQA

__all__ = ['BaseDumper', 'SafeDumper', 'Dumper', 'RoundTripDumper']


class BaseDumper(Emitter, Serializer, BaseRepresenter, BaseResolver):
    def __init__(
        self: Any,
        stream: StreamType,
        default_style: Any = None,
        default_flow_style: Any = None,
        canonical: Optional[bool] = None,
        indent: Optional[int] = None,
        width: Optional[int] = None,
        allow_unicode: Optional[bool] = None,
        line_break: Any = None,
        encoding: Any = None,
        explicit_start: Optional[bool] = None,
        explicit_end: Optional[bool] = None,
        version: Any = None,
        tags: Any = None,
        block_seq_indent: Any = None,
        top_level_colon_align: Any = None,
        prefix_colon: Any = None,
    ) -> None:
        # NOQA
        Emitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            allow_unicode=allow_unicode,
            line_break=line_break,
            block_seq_indent=block_seq_indent,
            dumper=self,
        )
        Serializer.__init__(
            self,
            encoding=encoding,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags,
            dumper=self,
        )
        BaseRepresenter.__init__(
            self,
            default_style=default_style,
            default_flow_style=default_flow_style,
            dumper=self,
        )
        BaseResolver.__init__(self, loadumper=self)


class SafeDumper(Emitter, Serializer, SafeRepresenter, Resolver):
    def __init__(
        self,
        stream: StreamType,
        default_style: Any = None,
        default_flow_style: Any = None,
        canonical: Optional[bool] = None,
        indent: Optional[int] = None,
        width: Optional[int] = None,
        allow_unicode: Optional[bool] = None,
        line_break: Any = None,
        encoding: Any = None,
        explicit_start: Optional[bool] = None,
        explicit_end: Optional[bool] = None,
        version: Any = None,
        tags: Any = None,
        block_seq_indent: Any = None,
        top_level_colon_align: Any = None,
        prefix_colon: Any = None,
    ) -> None:
        # NOQA
        Emitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            allow_unicode=allow_unicode,
            line_break=line_break,
            block_seq_indent=block_seq_indent,
            dumper=self,
        )
        Serializer.__init__(
            self,
            encoding=encoding,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags,
            dumper=self,
        )
        SafeRepresenter.__init__(
            self,
            default_style=default_style,
            default_flow_style=default_flow_style,
            dumper=self,
        )
        Resolver.__init__(self, loadumper=self)


class Dumper(Emitter, Serializer, Representer, Resolver):
    def __init__(
        self,
        stream: StreamType,
        default_style: Any = None,
        default_flow_style: Any = None,
        canonical: Optional[bool] = None,
        indent: Optional[int] = None,
        width: Optional[int] = None,
        allow_unicode: Optional[bool] = None,
        line_break: Any = None,
        encoding: Any = None,
        explicit_start: Optional[bool] = None,
        explicit_end: Optional[bool] = None,
        version: Any = None,
        tags: Any = None,
        block_seq_indent: Any = None,
        top_level_colon_align: Any = None,
        prefix_colon: Any = None,
    ) -> None:
        # NOQA
        Emitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            allow_unicode=allow_unicode,
            line_break=line_break,
            block_seq_indent=block_seq_indent,
            dumper=self,
        )
        Serializer.__init__(
            self,
            encoding=encoding,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags,
            dumper=self,
        )
        Representer.__init__(
            self,
            default_style=default_style,
            default_flow_style=default_flow_style,
            dumper=self,
        )
        Resolver.__init__(self, loadumper=self)


class RoundTripDumper(Emitter, Serializer, RoundTripRepresenter, VersionedResolver):
    def __init__(
        self,
        stream: StreamType,
        default_style: Any = None,
        default_flow_style: Optional[bool] = None,
        canonical: Optional[int] = None,
        indent: Optional[int] = None,
        width: Optional[int] = None,
        allow_unicode: Optional[bool] = None,
        line_break: Any = None,
        encoding: Any = None,
        explicit_start: Optional[bool] = None,
        explicit_end: Optional[bool] = None,
        version: Any = None,
        tags: Any = None,
        block_seq_indent: Any = None,
        top_level_colon_align: Any = None,
        prefix_colon: Any = None,
    ) -> None:
        # NOQA
        Emitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            allow_unicode=allow_unicode,
            line_break=line_break,
            block_seq_indent=block_seq_indent,
            top_level_colon_align=top_level_colon_align,
            prefix_colon=prefix_colon,
            dumper=self,
        )
        Serializer.__init__(
            self,
            encoding=encoding,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags,
            dumper=self,
        )
        RoundTripRepresenter.__init__(
            self,
            default_style=default_style,
            default_flow_style=default_flow_style,
            dumper=self,
        )
        VersionedResolver.__init__(self, loader=self)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import socket

from contrast import AGENT_CURR_WORKING_DIR
from contrast.utils.string_utils import truncate
from contrast_vendor.ruamel import yaml

from contrast.agent.protect.rule import ProtectionRule

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


CONFIG_LOCATIONS = [
    AGENT_CURR_WORKING_DIR,
    os.path.join(AGENT_CURR_WORKING_DIR, "settings"),
    "/etc/contrast/python/",
    "/etc/contrast/",
    "/etc/",
]

CONFIG_FILE_NAMES = ["contrast_security.yaml", "contrast_security.yml"]

# Valid options are defined in the spec:
# https://bitbucket.org/contrastsecurity/assess-specifications/src/master/vulnerability/capture-stacktrace.md
STACKTRACE_OPTIONS = ["ALL", "SOME", "NONE"]


def get_hostname():
    """
    Order of precedence for reporting server name:
    contrast_security.yaml (server.name) --> socket.gethostname() --> localhost
    """
    hostname = "localhost"

    try:
        hostname = socket.gethostname() or hostname
    except Exception as e:
        logger.debug(e)

    return truncate(hostname)


def load_yaml_config():
    """
    Checks for a yaml file at preconfigured file system location.

    See official documentation because this is valid across agents.

    Current order of precedence:
        file specified by CONTRAST_CONFIG_PATH env var
        os.getcwd()
        os.path.join(os.getcwd(), 'settings')
        /etc/contrast/python/
        /etc/contrast/
        /etc/

    :return: a dict object representation of the yaml config. {'enable': True, ....}
    """
    locations = CONFIG_LOCATIONS
    names = CONFIG_FILE_NAMES

    if "CONTRAST_CONFIG_PATH" in os.environ:
        filename = os.environ.get("CONTRAST_CONFIG_PATH")
        if os.path.isfile(filename):
            config = _load_config(filename)
            if config is not None:
                config["loaded_configuration_filename"] = filename
            return config
        logger.warning(
            "The path specified by CONTRAST_CONFIG_PATH is not a file -"
            " searching for configuration file in default locations",
            contrast_config_path=filename,
        )

    for path in locations:
        for name in names:
            file_path = os.path.join(path, name)

            if os.path.exists(file_path):
                # Common config dictates that agents should look only at the first
                # valid config file and not continue searching config files, even
                # if the first config cannot be loaded (due to format or else)
                config = _load_config(file_path)
                if config is not None:
                    config["loaded_configuration_filename"] = file_path
                return config

    return None


def _load_config(file_path):
    logger.info("Loading configuration file: %s", os.path.abspath(file_path))

    with open(file_path, "r") as config_file:
        try:
            return yaml.YAML(typ="safe", pure=True).load(config_file)
        except yaml.scanner.ScannerError as ex:
            # config yaml cannot be loaded but agent should continue on in case
            # env vars are configured
            msg_prefix = "YAML validator found an error."
            msg = f"{msg_prefix} Configuration path: [{ex.problem_mark.name}]. Line [{ex.problem_mark.line}]. Error: {ex.problem}"
            logger.warning(msg)

    return None


def flatten_config(config):
    """
    Convert a nested dict such as
        {'enable': True,
        'application':
            {'name': 'dani-flask'},
        'agent':
            {'python':
                {'enable_rep': True}...

    into
        {'enable': True,
        'application.name': 'dani-flask',
        'agent.python.enable_rep': True,

    :param config: dict config with nested keys and values
    :return: dict, flattened where each key has one value.
    """
    flattened_config = {}

    def flatten(x, name=""):
        if isinstance(x, dict):
            for key in x:
                flatten(x[key], name + key + ".")
        else:
            flattened_config[name[:-1]] = x

    flatten(config)
    return flattened_config


def str_to_bool(val):
    """
    Converts a str to a bool

    true -> True, false -> False
    """
    if isinstance(val, bool):
        return val
    if not isinstance(val, str):
        return False

    # The remainder of this function was ported from distutils
    val = val.lower()

    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    if val in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError(f"invalid truth value {val!r}")


def get_env_key(key):
    return os.environ.get(key)


def parse_disabled_rules(disabled_rules):
    if not disabled_rules:
        return []

    return [rule.lower() for rule in disabled_rules.split(",")]


def parse_stacktraces_options(option):
    option = option.upper()
    if option in STACKTRACE_OPTIONS:
        return option

    return "ALL"


def str_to_protect_mode_enum(mode):
    """
    Converts str config value to protect mode enum that the agent understands
    """
    if not mode:
        return ProtectionRule.OFF

    return getattr(ProtectionRule, mode.upper(), ProtectionRule.MONITOR)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re
from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction
from contrast.utils.decorators import fail_quietly

# Used to determine if the input violates the constraint in the
# specification linked above, meaning it has two or more layers of nested
# multi-match groups.
REDOS_REGEX_STRING = r"[\[(].*?[\[(].*?[\])][*+?].*?[\])][*+?]"
# For matching bytes and bytearray objects
REDOS_REGEX_BYTES = REDOS_REGEX_STRING.encode()


# re pattern type varies across Py versions and some even have bugs
# https://github.com/beetbox/beets/issues/2986
# So while this looks hacky, it will work.
PATTERN_TYPE = type(re.compile(""))


class RedosAction(DefaultAction):
    """
    Custom trigger action that implements ReDos rule logic.

    A ReDOS vulnerability is triggered when the user input is untrusted (as
    with other rules) but unlike other rules, when the regex passed in to the re
    method matches a specific pattern. This rule deviates from other Assess rules in
    that the value provided may change the outcome of the vulnerability detection.

    https://bitbucket.org/contrastsecurity/assess-specifications/src/master/rules/dataflow/redos.md
    """

    def is_violated(
        self, source_properties, required_tags, disallowed_tags, orig_args=None
    ):
        # 1) Check that user input is untrusted.
        violated = super().is_violated(
            source_properties, required_tags, disallowed_tags
        )
        if not violated:
            return False

        # 2) check that regex passed in to re method matches our ReDOs regex.
        return self._is_complex_regex(orig_args)

    @fail_quietly("Failed check regex complexity", return_value=False)
    def _is_complex_regex(self, orig_args):
        if not orig_args:
            return False

        # orig_args is (regex, user_string)
        regex = orig_args[0]

        if isinstance(regex, PATTERN_TYPE):
            regex = regex.pattern

        redos_regex = (
            REDOS_REGEX_STRING if isinstance(regex, str) else REDOS_REGEX_BYTES
        )

        # Even though Ruby uses `match`, we have to use `search` which checks for
        # a match anywhere in the string, not just the beginning.
        return bool(re.search(redos_regex, regex, flags=re.IGNORECASE))



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import ast
import textwrap
import functools

from collections import deque
from copy import copy
from importlib import import_module
from types import FunctionType, MethodType
from inspect import isclass, getsource
from django.urls.exceptions import Resolver404

from django.utils.regex_helper import normalize

try:
    from django.urls import get_resolver
except ImportError:
    from django.core.urlresolvers import get_resolver

from contrast.agent.middlewares.route_coverage.common import (
    DEFAULT_ROUTE_METHODS,
    build_args_from_function,
    build_key,
)
from contrast.api import Route
from contrast.agent.settings import Settings
from functools import lru_cache

from contrast_vendor import structlog as logging
from contrast.utils.decorators import fail_quietly

logger = logging.getLogger("contrast")


@lru_cache(maxsize=1)
def is_django_1():
    settings = Settings(framework_name="django")
    return settings.framework.version.major == "1"


def check_for_http_decorator(func):
    """
    Grabs the require_http_methods decorator from a view function call via inspect

    NOTE: this only works for require_http_methods; it does not currently work for require_GET,
    require_POST, require_safe
    """
    method_types = {}

    if isinstance(func, functools.partial):
        func = getattr(func, "func", None)

    if func is None:
        return method_types

    def visit_function_def(node):
        for decorator in node.decorator_list:
            if isinstance(decorator, ast.Call):
                # decorator name which should be require_http_methods
                if getattr(decorator.func, "id", "") == "require_http_methods":
                    name = (
                        decorator.func.attr
                        if isinstance(decorator.func, ast.Attribute)
                        else decorator.func.id
                    )
                    method_list = decorator.args[0] if len(decorator.args) > 0 else []
                    method_str_list = list(
                        filter(
                            None,
                            [
                                getattr(s, "s", "")  # py37: change "s" to "value"
                                for s in getattr(method_list, "elts", [])
                            ],
                        )
                    )
                    if not method_str_list:
                        continue
                    method_types[name] = method_str_list
                    return

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_function_def
    node_source = textwrap.dedent(getsource(func))
    node_iter.visit(ast.parse(node_source))

    return method_types


def get_lowest_function_call(func):
    if isclass(func) or func.__closure__ is None:
        return func
    closure = (c.cell_contents for c in func.__closure__)
    return next((c for c in closure if isinstance(c, (FunctionType, MethodType))), None)


def create_url(pattern_or_resolver):
    if is_django_1():
        pattern = pattern_or_resolver.regex.pattern
    else:
        pattern = pattern_or_resolver.pattern.regex.pattern

    try:
        normalized = normalize(pattern)[0][0]
        url = normalized.replace("%(", "{").replace(")", "}")
    except Exception:
        url = pattern_or_resolver.name

    return url


def get_method_info(pattern_or_resolver):
    method_types = []
    method_arg_names = "()"

    lowest_function = get_lowest_function_call(pattern_or_resolver.callback)

    if lowest_function is not None:
        method_arg_names = build_args_from_function(lowest_function)

        # this method returns a dict because it uses one to store state in the recursive function
        method_types_dict = check_for_http_decorator(lowest_function)
        method_types = method_types_dict.get("require_http_methods", [])

        if not isinstance(method_types, list):
            method_types = [method_types]

    return method_types or DEFAULT_ROUTE_METHODS, method_arg_names


def create_one_route(routes, method_type, method_arg_names, pattern_or_resolver):
    """
    Create a new api.Route object and adds it to routes dict

    :param routes: dict of routes
    :param method_type:
    :param method_arg_names:
    :param pattern_or_resolver:
    :return: None, adds to routes dict which is updated through pass-by-reference
    """
    route = build_django_route(pattern_or_resolver, method_arg_names)

    route_id = str(id(pattern_or_resolver.callback))

    key = build_key(route_id, method_type)

    routes[key] = Route(
        verb=method_type, url=create_url(pattern_or_resolver), route=route
    )


def create_routes(urlpatterns):
    if is_django_1():
        try:
            # Django 1.10 and above
            from django.urls.resolvers import RegexURLPattern, RegexURLResolver
        except ImportError:
            # Django 1.9 and below
            from django.core.urlresolvers import RegexURLPattern, RegexURLResolver
    else:
        from django.urls.resolvers import (
            URLPattern as RegexURLPattern,
            URLResolver as RegexURLResolver,
        )

    routes = {}

    urlpatterns_deque = deque(urlpatterns)

    while urlpatterns_deque:
        url_pattern = urlpatterns_deque.popleft()

        if isinstance(url_pattern, RegexURLResolver):
            urlpatterns_deque.extend(url_pattern.url_patterns)

        elif isinstance(url_pattern, RegexURLPattern):
            method_types, method_arg_names = get_method_info(url_pattern)
            for method_type in method_types:
                create_one_route(routes, method_type, method_arg_names, url_pattern)
    return routes


def create_django_routes():
    """
    Grabs all URL's from the root settings and searches for possible required_method decorators

    In Django there is no implicit declaration of GET or POST. Often times decorators are used to fix this.

    Returns a dict of key = id, value = api.Route.
    """

    from django.conf import settings

    if not settings.ROOT_URLCONF:
        logger.info("Application does not define settings.ROOT_URLCONF")
        logger.debug("Skipping enumeration of urlpatterns")
        return None

    try:
        root_urlconf = import_module(settings.ROOT_URLCONF)
    except Exception as exception:
        logger.debug("Failed to import ROOT_URLCONF: %s", exception)
        return None

    try:
        urlpatterns = root_urlconf.urlpatterns or []
    except Exception as exception:
        logger.debug("Failed to get urlpatterns: %s", exception)
        return None

    url_patterns = copy(urlpatterns)
    return create_routes(url_patterns)


def _function_loc(func):
    """Return the function's module and name"""
    return f"{func.__module__}.{func.__name__}"


def build_django_route(obj, method_arg_names=None):
    if hasattr(obj, "lookup_str"):
        route = obj.lookup_str
    elif hasattr(obj, "callback"):
        cb = obj.callback
        route = _function_loc(cb)
    elif callable(obj):
        route = _function_loc(obj)
    else:
        logger.debug("WARNING: can't build django route for object type %s", type(obj))
        return ""

    if method_arg_names is None:
        method_arg_names = build_args_from_function(obj)

    route += method_arg_names
    return route


@fail_quietly("Failed to get view function for django application")
def get_view_func(path):
    from django.conf import settings

    try:
        result = get_resolver().resolve(path or "/")
    except Resolver404:
        return None

    if (
        result is None
        and not path.endswith("/")
        and "django.middleware.common.CommonMiddleware" in settings.MIDDLEWARE
        and settings.APPEND_SLASH
    ):
        result = get_view_func(f"{path}/")
    if result is None:
        return None

    return result.func


class DjangoRoutesMixin:
    @fail_quietly("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        """
        Route Coverage is the Assess feature that looks for routes generally defined
        in Django apps in a file like urls.py
        """
        return create_django_routes()

    @fail_quietly("Unable to get Django view func")
    def get_view_func(self, request):
        return get_view_func(self.request_path)

    @fail_quietly("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_django_route(view_func)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_source_nodes
from contrast.agent.policy.utils import CompositeNode

django_sources = [
    CompositeNode(
        {
            "module": "django.http.request",
            "class_name": "HttpRequest",
        },
        [
            {
                "method_name": [
                    "get_host",
                    "get_port",
                    "get_full_path",
                    "get_full_path_info",
                    "get_raw_uri",
                    "build_absolute_uri",
                ],
                "node_type": "URI",
                "tags": ["CROSS_SITE"],
            },
            {
                "method_name": ["body", "read"],
                "node_type": "BODY",
                "tags": ["CROSS_SITE"],
            },
            {
                "method_name": ["scheme", "encoding"],
                "node_type": "OTHER",
                "tags": ["CROSS_SITE"],
            },
        ],
    ),
    CompositeNode(
        {
            "module": "django.http.request",
            "class_name": "HttpHeaders",
        },
        [
            {
                # Header can't trigger XSS unless the key is Referer
                "method_name": "__getitem__",
                "node_type": "HEADER",
                "tags": ["NO_NEWLINES"],
            },
            {
                # Header key can't trigger XSS
                "method_name": "keys",
                "node_type": "HEADER_KEY",
                "tags": ["NO_NEWLINES"],
            },
        ],
    ),
    {
        "module": "django.utils.datastructures",
        "class_name": "MultiValueDict",
        "method_name": "__getitem__",
        "node_type": "PARAMETER",
        "tags": ["CROSS_SITE"],
    },
    {
        # This was added to handle file uploads in Django REST Framework.
        # For regular Django, this is redundant with the file tracker in middleware
        "module": "django.core.files.uploadedfile",
        "class_name": "UploadedFile",
        "method_name": "__init__",
        "target": "ARG_0,KWARG:file",
        "node_type": "MULTIPART_CONTENT_DATA",
        "tags": ["CROSS_SITE"],
    },
]


register_source_nodes(django_sources)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class Heartbeat(BaseTsAppMessage):
    def __init__(self):
        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"

    @property
    def name(self):
        return "heartbeat"

    @property
    def path(self):
        return "/".join(
            [
                "applications",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                "heartbeat",
            ]
        )

    @property
    def request_method(self):
        return requests.post

    @fail_loudly("Failed to process Heartbeat response")
    def process_response(self, response, reporting_client):
        """
        This endpoint should only ever return a statuscode and empty body.
        Only in the case of a silencing status code, should we take an action.
        """
        self.process_response_code(response, reporting_client)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.timer import now_ms


class Sample:
    def __init__(self):
        self.timestamp_ms = now_ms()
        self.user_input = None
        self.stack_trace_elements = []
        self.details = {}

    def set_stack(self, stack):
        self.stack_trace_elements = stack

    def set_user_input(self, user_input):
        self.user_input = user_input



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_source_nodes
from contrast.agent.policy.utils import CompositeNode


asgi_sources = [
    {
        "module": "starlette.requests",
        "class_name": "Request",
        "method_name": "body",
        "node_type": "BODY",
        "tags": ["CROSS_SITE"],
    },
    CompositeNode(
        {
            "module": "starlette.datastructures",
            "class_name": "FormData",
        },
        [
            {
                # We need get in addition to __getitem__, because starlette overrides them
                # both, and one doesn't call the other.
                "method_name": ["__getitem__", "get"],
                # MULTIPART_FORM_DATA isn't totally accurate, because forms could use
                # other content types
                "node_type": "MULTIPART_FORM_DATA",
                "tags": ["CROSS_SITE"],
            },
        ],
    ),
    {
        "module": "yarl",
        "class_name": "URL",
        "method_name": "__str__",
        "node_type": "URI",
        "tags": ["CROSS_SITE"],
    },
    CompositeNode(
        {
            "module": "aiohttp.web_request",
            "class_name": "BaseRequest",
        },
        [
            {
                "method_name": ["read", "json", "post"],
                "node_type": "BODY",
                "tags": ["CROSS_SITE"],
            },
        ],
    ),
    {
        "module": "aiohttp.multipart",
        "class_name": "BodyPartReader",
        "method_name": "read",
        "node_type": "MULTIPART_FORM_DATA",
        "tags": ["CROSS_SITE"],
    },
]


register_source_nodes(asgi_sources)



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

from __future__ import annotations

import sys
import traceback

from io import StringIO
from types import FrameType

from .typing import ExcInfo


def _format_exception(exc_info: ExcInfo) -> str:
    """
    Prettyprint an `exc_info` tuple.

    Shamelessly stolen from stdlib's logging module.
    """
    sio = StringIO()

    traceback.print_exception(exc_info[0], exc_info[1], exc_info[2], None, sio)
    s = sio.getvalue()
    sio.close()
    if s[-1:] == "\n":
        s = s[:-1]

    return s


def _find_first_app_frame_and_name(
    additional_ignores: list[str] | None = None,
) -> tuple[FrameType, str]:
    """
    Remove all intra-structlog calls and return the relevant app frame.

    :param additional_ignores: Additional names with which the first frame must
        not start.

    :returns: tuple of (frame, name)
    """
    ignores = ["structlog"] + (additional_ignores or [])
    f = sys._getframe()
    name = f.f_globals.get("__name__") or "?"
    while any(tuple(name.startswith(i) for i in ignores)):
        if f.f_back is None:
            name = "?"
            break
        f = f.f_back
        name = f.f_globals.get("__name__") or "?"
    return f, name


def _format_stack(frame: FrameType) -> str:
    """
    Pretty-print the stack of *frame* like logging would.
    """
    sio = StringIO()

    sio.write("Stack (most recent call last):\n")
    traceback.print_stack(frame, file=sio)
    sinfo = sio.getvalue()
    if sinfo[-1] == "\n":
        sinfo = sinfo[:-1]
    sio.close()

    return sinfo



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.response.base_response_rule import BaseResponseRule
from contrast.api import Finding, send_finding


class CacheControlsRule(BaseResponseRule):
    GOOD_VALUES = ("no-cache", "no-store")

    @property
    def name(self):
        return "cache-controls-missing"

    def is_violated(self, headers, _, form_tags, meta_tags):
        header_violated, header_properties = self.is_header_violated(headers)
        if not header_violated:
            # If header is correctly configured for this rule, no need to check body.
            return False, {}

        body_violated, body_properties = self.is_body_violated(meta_tags)
        if not body_violated:
            # Even though header may have violated the rule, if the body was configured
            # correctly for this rule, the rule is not violated.
            return False, {}

        return True, header_properties or body_properties

    def is_header_violated(self, headers):
        """
        Rule is violated if Cache-Control header is present OR it is present
        but not assigned to a desired value.
        :param headers: dict of response headers
        :return: (bool, properties)
        """
        value = headers.get("Cache-Control")
        if not value:
            # Finding gets no properties if header is missing
            return True, {}
        if any(v in value for v in self.GOOD_VALUES):
            return False, None

        properties = dict(
            type="Header",
            name="Cache-Control",
            value=value,
        )
        return True, properties

    def is_body_violated(self, meta_tags):
        """
        Rule is violated if:
        1. none of the attrs of the list of form tags have an "autocomplete" attr
        2. at least one attr is "autocomplete" but it is is assigned to anything other than "off"
        :param body: response body
        :param form_tags: list of Tag namedtuple
        :return: bool, properties dict
        """
        if not meta_tags:
            return True, {}

        first_relevant_tag = None
        for tag in meta_tags:
            for attr_idx, (attr_name, attr_value) in enumerate(tag.attrs):
                if (
                    attr_name.lower() == "http-equiv"
                    and attr_value.lower() == "cache-control"
                ):
                    first_relevant_tag = tag
                    content, value = tag.attrs[attr_idx + 1]
                    if content.lower() == "content" and any(
                        v in value.lower() for v in self.GOOD_VALUES
                    ):
                        return False, {}

        # If no cache-control attr in any meta tag, report the first meta tag since
        # they are all in violation.
        properties = self.build_properties(first_relevant_tag)
        return True, properties

    def build_properties(self, tag):
        if not tag:
            return dict()
        return dict(
            type=tag.type,
            name="pragma",
            value=tag.tag,
        )

    def create_finding(self, properties):
        if not properties:
            data = dict(data="[]")
        else:
            data = dict(data=f"[{str(properties)}]")
        return Finding(self, data)

    def build_and_append_finding(self, properties, context):
        finding = self.create_finding(properties)
        send_finding(finding, context)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.agent.assess.utils import copy_from


class PrependPropagator(BasePropagator):
    def _propagate(self):
        try:
            original_start_index = self.target.rindex(self.first_source)
        except ValueError:
            original_start_index = 0

        # find original in the target, copy tags to the new position in target
        copy_from(
            self.target, self.first_source, original_start_index, self.node.untags
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import time


def now_ms_float():
    return time.time() * 1000


def now_ms():
    return int(now_ms_float())


def sleep(seconds):
    """
    Wrapper around time.sleep that makes testing less complicated
    """
    time.sleep(seconds)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.trigger_rule import TriggerRule


class NonDataflowRule(TriggerRule):
    """
    Rule class used for all non-dataflow rules (i.e. crypto rules).
    """

    def update_preflight_hash(self, hasher, source=None, **kwargs):
        """
        This gets called from `create_finding` in the base class
        """
        # source corresponds to the algorithm name for crytpo rules
        if source is not None:
            hasher.update(source)

    @classmethod
    def from_nodes(cls, name: str, nodes):
        return super().from_nodes(name, False, nodes)



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Extract a structured traceback from an exception.

Contributed by Will McGugan (see
https://github.com/hynek/structlog/pull/407#issuecomment-1150926246) from Rich:
https://github.com/Textualize/rich/blob/972dedff/rich/traceback.py
"""

from __future__ import annotations

import os

from dataclasses import asdict, dataclass, field
from traceback import walk_tb
from types import TracebackType
from typing import Any, Tuple, Union

from .typing import ExcInfo


__all__ = [
    "ExceptionDictTransformer",
    "Frame",
    "Stack",
    "SyntaxError_",
    "Trace",
    "extract",
    "safe_str",
    "to_repr",
]


SHOW_LOCALS = True
LOCALS_MAX_STRING = 80
MAX_FRAMES = 50

OptExcInfo = Union[ExcInfo, Tuple[None, None, None]]


@dataclass
class Frame:
    """
    Represents a single stack frame.
    """

    filename: str
    lineno: int
    name: str
    line: str = ""
    locals: dict[str, str] | None = None


@dataclass
class SyntaxError_:
    """
    Contains detailed information about :exc:`SyntaxError` exceptions.
    """

    offset: int
    filename: str
    line: str
    lineno: int
    msg: str


@dataclass
class Stack:
    """
    Represents an exception and a list of stack frames.
    """

    exc_type: str
    exc_value: str
    syntax_error: SyntaxError_ | None = None
    is_cause: bool = False
    frames: list[Frame] = field(default_factory=list)


@dataclass
class Trace:
    """
    Container for a list of stack traces.
    """

    stacks: list[Stack]


def safe_str(_object: Any) -> str:
    """Don't allow exceptions from __str__ to propegate."""
    try:
        return str(_object)
    except Exception as error:
        return f"<str-error {str(error)!r}>"


def to_repr(obj: Any, max_string: int | None = None) -> str:
    """Get repr string for an object, but catch errors."""
    if isinstance(obj, str):
        obj_repr = obj
    else:
        try:
            obj_repr = repr(obj)
        except Exception as error:
            obj_repr = f"<repr-error {str(error)!r}>"

    if max_string is not None and len(obj_repr) > max_string:
        truncated = len(obj_repr) - max_string
        obj_repr = f"{obj_repr[:max_string]!r}+{truncated}"

    return obj_repr


def extract(
    exc_type: type[BaseException],
    exc_value: BaseException,
    traceback: TracebackType | None,
    *,
    show_locals: bool = False,
    locals_max_string: int = LOCALS_MAX_STRING,
) -> Trace:
    """
    Extract traceback information.

    :param exc_type: Exception type.
    :param exc_value: Exception value.
    :param traceback: Python Traceback object.
    :param show_locals: Enable display of local variables. Defaults to False.
    :param locals_max_string: Maximum length of string before truncating, or
        ``None`` to disable.
    :param max_frames: Maximum number of frames in each stack

    :returns: A Trace instance with structured information about all
        exceptions.

    .. versionadded:: 22.1
    """

    stacks: list[Stack] = []
    is_cause = False

    while True:
        stack = Stack(
            exc_type=safe_str(exc_type.__name__),
            exc_value=safe_str(exc_value),
            is_cause=is_cause,
        )

        if isinstance(exc_value, SyntaxError):
            stack.syntax_error = SyntaxError_(
                offset=exc_value.offset or 0,
                filename=exc_value.filename or "?",
                lineno=exc_value.lineno or 0,
                line=exc_value.text or "",
                msg=exc_value.msg,
            )

        stacks.append(stack)
        append = stack.frames.append  # pylint: disable=no-member

        for frame_summary, line_no in walk_tb(traceback):
            filename = frame_summary.f_code.co_filename
            if filename and not filename.startswith("<"):
                filename = os.path.abspath(filename)
            frame = Frame(
                filename=filename or "?",
                lineno=line_no,
                name=frame_summary.f_code.co_name,
                locals={
                    key: to_repr(value, max_string=locals_max_string)
                    for key, value in frame_summary.f_locals.items()
                }
                if show_locals
                else None,
            )
            append(frame)

        cause = getattr(exc_value, "__cause__", None)
        if cause and cause.__traceback__:
            exc_type = cause.__class__
            exc_value = cause
            traceback = cause.__traceback__
            is_cause = True
            continue

        cause = exc_value.__context__
        if (
            cause
            and cause.__traceback__
            and not getattr(exc_value, "__suppress_context__", False)
        ):
            exc_type = cause.__class__
            exc_value = cause
            traceback = cause.__traceback__
            is_cause = False
            continue

        # No cover, code is reached but coverage doesn't recognize it.
        break  # pragma: no cover

    return Trace(stacks=stacks)


class ExceptionDictTransformer:
    """
    Return a list of exception stack dictionaries for an exception.

    These dictionaries are based on :class:`Stack` instances generated by
    :func:`extract()` and can be dumped to JSON.

    :param show_locals: Whether or not to include the values of a stack frame's
        local variables.
    :param locals_max_string: The maximum length after which long string
        representations are truncated.
    :param max_frames: Maximum number of frames in each stack.  Frames are
        removed from the inside out.  The idea is, that the first frames
        represent your code responsible for the exception and last frames the
        code where the exception actually happened.  With larger web
        frameworks, this does not always work, so you should stick with the
        default.
    """

    def __init__(
        self,
        show_locals: bool = True,
        locals_max_string: int = LOCALS_MAX_STRING,
        max_frames: int = MAX_FRAMES,
    ) -> None:
        if locals_max_string < 0:
            raise ValueError(
                f'"locals_max_string" must be >= 0: {locals_max_string}'
            )
        if max_frames < 2:
            raise ValueError(f'"max_frames" must be >= 2: {max_frames}')
        self.show_locals = show_locals
        self.locals_max_string = locals_max_string
        self.max_frames = max_frames

    def __call__(self, exc_info: ExcInfo) -> list[dict[str, Any]]:
        trace = extract(
            *exc_info,
            show_locals=self.show_locals,
            locals_max_string=self.locals_max_string,
        )

        for stack in trace.stacks:
            if len(stack.frames) <= self.max_frames:
                continue

            half = (
                self.max_frames // 2
            )  # Force int division to handle odd numbers correctly
            fake_frame = Frame(
                filename="",
                lineno=-1,
                name=f"Skipped frames: {len(stack.frames) - (2 * half)}",
            )
            stack.frames[:] = [
                *stack.frames[:half],
                fake_frame,
                *stack.frames[-half:],
            ]

        return [asdict(stack) for stack in trace.stacks]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast_vendor.wrapt import register_post_import_hook

import contrast
from contrast.agent import scope
from contrast.agent.policy.applicator import apply_module_patches
from contrast.agent.middlewares.route_coverage.common import build_route
from contrast.agent.middlewares.route_coverage.bottle_routes import create_bottle_routes
from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark
from contrast.utils.decorators import fail_quietly

from contrast_vendor import structlog as logging

MODULE_NAME = "bottle"

logger = logging.getLogger("contrast")


def build_prepare_patch(orig_func, patch_policy, bottle_module):
    del patch_policy

    def prepare_patch(wrapped, instance, args, kwargs):
        """
        Patch for bottle.SimpleTemplate.prepare.

        This is needed because of the unfortunate way bottle calls on
        `html_escape` as a kwarg in the prepare definition in SimpleTemplate.
        See https://github.com/bottlepy/bottle/blob/master/bottle.py#L3952

        Because of this behavior, the `html_escape` func is not our patched
        `html_escape` defined in policy.
        By patching prepare, we intercept its call and instead of allowing
        it to use the default kwarg for `html_escape`, we pass our own
        patched `html_escape` in order to prevent false positive XSS findings.
        """
        del instance

        # html_escape MUST already be patched by policy in order
        # to pass in the patched func to prepare
        kwargs.setdefault("escape_func", bottle_module.html_escape)
        return wrapped(*args, **kwargs)

    return wrap_and_watermark(orig_func, prepare_patch)


def build_call_patch(orig_func, patch_policy):
    """
    Patch for bottle.Bottle.__call__

    This is how we perform route discovery for Bottle. Many different patches could
    probably work here, since we really just need a reference to the Bottle instance
    after all routes have been registered.
    """
    del patch_policy

    def call_patch(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)
        do_bottle_route_discovery(instance)
        return result

    return wrap_and_watermark(orig_func, call_patch)


@fail_quietly("unable to perform Bottle route discovery")
def do_bottle_route_discovery(bottle_instance):
    with scope.contrast_scope():
        from contrast.agent import agent_state

        if not agent_state.is_first_request():
            return

        discovered_routes = create_bottle_routes(bottle_instance)
        agent_state.update_routes(discovered_routes)
        logger.debug(
            "Discovered the following Bottle routes: %s",
            [f"{route.verb} {route.url}" for route in discovered_routes.values()],
        )


def build_match_patch(orig_func, patch_policy):
    """
    Patch for bottle.Router.match()

    This sets up request context with all necessary info for current route observation
    """
    del patch_policy

    def match_patch(wrapped, instance, args, kwargs):
        del instance

        result = wrapped(*args, **kwargs)
        do_bottle_route_observation(result)
        return result

    return wrap_and_watermark(orig_func, match_patch)


@fail_quietly("unable to perform bottle route coverage in Bottle match patch")
def do_bottle_route_observation(match_result):
    with scope.contrast_scope():
        if match_result is None:
            return

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            return

        context.view_func = match_result[0].callback
        context.view_func_str = build_route(
            context.request.get_normalized_uri(),
            context.view_func,
        )
        logger.debug("Observed Bottle route: %s", context.view_func_str)


def patch_bottle(bottle_module):
    # We ask policy to go ahead and do all bottle patches here (even though policy
    # patches will happen later on) because we MUST have some bottle policy patches
    # already applied for these non-policy patches to work.
    # This would not be necessary if in _enable_patches policy_patches were applied
    # first.
    apply_module_patches(bottle_module)

    build_and_apply_patch(
        bottle_module.SimpleTemplate,
        "prepare",
        build_prepare_patch,
        builder_args=(bottle_module,),
    )
    build_and_apply_patch(bottle_module.Bottle, "__call__", build_call_patch)
    build_and_apply_patch(bottle_module.Router, "match", build_match_patch)


def register_patches():
    register_post_import_hook(patch_bottle, MODULE_NAME)


def reverse_patches():
    bottle_module = sys.modules.get(MODULE_NAME)
    if not bottle_module:
        return

    patch_manager.reverse_patches_by_owner(bottle_module)
    patch_manager.reverse_patches_by_owner(bottle_module.SimpleTemplate)
    patch_manager.reverse_patches_by_owner(bottle_module.Bottle)
    patch_manager.reverse_patches_by_owner(bottle_module.Router)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from contrast.agent.policy import constants
from contrast.agent.assess.assess_exceptions import ContrastAssessException
from contrast.agent.policy.policy_node import PolicyNode
from contrast.api.trace_event import TraceEvent


class SourceNode(PolicyNode):
    ALL = "ALL"

    SOURCE_TAG = "UNTRUSTED"

    def __init__(
        self,
        module,
        class_name="",
        instance_method=True,
        method_name="",
        target="RETURN",
        node_type=None,
        tags=None,
        source_name="",
        version=ALL,
        policy_patch=True,
    ):
        super().__init__(
            module,
            class_name,
            instance_method,
            method_name,
            source_name,
            target,
            tags,
            policy_patch=policy_patch,
        )

        self.version = version
        self.type = node_type

        if tags is None:
            self.tags = {self.SOURCE_TAG}
        else:
            self.tags.add(self.SOURCE_TAG)

        self.validate()

    @classmethod
    def dynamic(cls, module, class_name, column_name, tags, policy_patch=True):
        """
        Create a dynamic source node, one with the dynamic_source_id property.
        """
        instance_method = True
        method_name = column_name
        target = "RETURN"

        node = cls(
            module,
            class_name,
            instance_method,
            method_name,
            target,
            constants.DB_SOURCE_TYPE,
            tags=tags,
            policy_patch=policy_patch,
        )

        node.add_property("dynamic_source_id", node.name)

        return node

    @property
    def node_type(self):
        """
        This is confusing. Sources are Creation action but Propagation type. Oh
        and also Type refers to input type, like parameter, so we have to call
        this node_type. :-/
        """
        return TraceEvent.EventType.PROPAGATION

    def validate(self):
        super().validate()

        if not self.targets:
            raise ContrastAssessException(
                f"Source {self.name} did not have a proper target. Unable to create."
            )

        if not self.type:
            raise ContrastAssessException(
                f"Source {self.name} did not have a proper type. Unable to create."
            )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class SettingsREPMixin(object):
    valid_rep_features = {
        "cmd-injection": [
            "detect_parameter_command_backdoors",
            "detect_chained_commands",
        ],
        "path-traversal": [
            "detect_custom_code_accessing_system_files",
            "detect_common_file_exploits",
        ],
        "sql-injection": [
            "detect_tautologies",
            "detect_dangerous_functions",
            "detect_chained_queries",
            "detect_suspicious_unions",
        ],
        "malformed-header": ["mode"],
    }

    def is_rep_feature_enabled_for_rule(self, rule_name, feature):
        rep_feature = ".".join(["protect", "rules", rule_name, feature])
        return self.config.get(rep_feature, True)

    @staticmethod
    def is_valid_rep_feature_for_rule(rule_name, key):
        if not key or not rule_name:
            return False
        return key in SettingsREPMixin.valid_rep_features[rule_name]

    @staticmethod
    def is_valid_value_for_rep_feature(value):
        return isinstance(value, bool)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.sql_injection.default_sql_scanner import (
    DefaultSqlScanner,
)


class MongoNoSqlScanner(DefaultSqlScanner):
    LEFT_ANGLE = "<"

    def start_line_comment(self, char, index, query):
        if char == self.SLASH_CHAR and query[index + 1] == self.SLASH_CHAR:
            return True

        if (
            char == self.LEFT_ANGLE
            and query[index + 1] == self.DASH_CHAR
            and query[index + 2] == self.DASH_CHAR
        ):
            return True

        return False

    def start_block_comment(self, char, index, query):
        return False

    def double_quote_escape_in_double_quote(self):
        return True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from importlib import import_module


def safe_import_list(*import_names: str):
    return list(filter(None, [safe_import(x) for x in import_names]))


def safe_import(full_name: str):
    module_name, import_name = full_name.rsplit(".", 1)

    try:
        module = import_module(module_name)
        return getattr(module, import_name)
    except (AttributeError, ModuleNotFoundError):
        return None
